@@ -23,7 +23,7 @@
  * @TYPE_INET_SERVICE:	internet service (integer subtype)
  * @TYPE_ICMP_TYPE:	ICMP type codes (integer subtype)
  * @TYPE_TCP_FLAG:	TCP flag (bitmask subtype)
- * @TCPE_DCCP_PKTTYPE:	DCCP packet type (integer subtype)
+ * @TYPE_DCCP_PKTTYPE:	DCCP packet type (integer subtype)
  * @TYPE_MH_TYPE:	Mobility Header type (integer subtype)
  * @TYPE_TIME:		relative time
  * @TYPE_MARK:		packet mark (integer subtype)
@@ -152,9 +152,7 @@ struct parse_ctx;
  */
 struct datatype {
 	uint32_t			type;
-	enum byteorder			byteorder:8;
-	uint32_t			alloc:1,
-					is_typeof:1;
+	enum byteorder			byteorder;
 	unsigned int			flags;
 	unsigned int			size;
 	unsigned int			subtypes;
@@ -178,7 +176,6 @@ extern const struct datatype *datatype_lookup(enum datatypes type);
 extern const struct datatype *datatype_lookup_byname(const char *name);
 extern struct datatype *datatype_get(const struct datatype *dtype);
 extern void datatype_set(struct expr *expr, const struct datatype *dtype);
-extern void __datatype_set(struct expr *expr, const struct datatype *dtype);
 extern void datatype_free(const struct datatype *dtype);
 struct datatype *dtype_clone(const struct datatype *orig_dtype);
 
@@ -280,12 +277,6 @@ extern const struct datatype boolean_type;
 extern const struct datatype priority_type;
 extern const struct datatype policy_type;
 extern const struct datatype cgroupv2_type;
-extern const struct datatype queue_type;
-
-/* private datatypes for reject statement. */
-extern const struct datatype reject_icmp_code_type;
-extern const struct datatype reject_icmpv6_code_type;
-extern const struct datatype reject_icmpx_code_type;
 
 void inet_service_type_print(const struct expr *expr, struct output_ctx *octx);
 
