@@ -62,7 +62,6 @@ static const struct datatype *datatypes[TYPE_MAX + 1] = {
 	[TYPE_CT_DIR]		= &ct_dir_type,
 	[TYPE_CT_STATUS]	= &ct_status_type,
 	[TYPE_ICMP6_TYPE]	= &icmp6_type_type,
-	[TYPE_CT_LABEL]		= &ct_label_type,
 	[TYPE_PKTTYPE]		= &pkttype_type,
 	[TYPE_ICMP_CODE]	= &icmp_code_type,
 	[TYPE_ICMPV6_CODE]	= &icmpv6_code_type,
@@ -72,7 +71,6 @@ static const struct datatype *datatypes[TYPE_MAX + 1] = {
 	[TYPE_ECN]		= &ecn_type,
 	[TYPE_FIB_ADDR]         = &fib_addr_type,
 	[TYPE_BOOLEAN]		= &boolean_type,
-	[TYPE_CT_EVENTBIT]	= &ct_event_type,
 	[TYPE_IFNAME]		= &ifname_type,
 	[TYPE_IGMP_TYPE]	= &igmp_type_type,
 	[TYPE_TIME_DATE]	= &date_type,
@@ -416,24 +414,6 @@ const struct datatype xinteger_type = {
 	.parse		= integer_type_parse,
 };
 
-static void queue_type_print(const struct expr *expr, struct output_ctx *octx)
-{
-	nft_gmp_print(octx, "queue");
-}
-
-/* Dummy queue_type for set declaration with typeof, see
- * constant_expr_build_udata and constant_expr_parse_udata,
- * basetype is used for elements.
-*/
-const struct datatype queue_type = {
-	.type		= TYPE_INTEGER,
-	.is_typeof	= 1,
-	.name		= "queue",
-	.desc		= "queue",
-	.basetype	= &integer_type,
-	.print		= queue_type_print,
-};
-
 static void string_type_print(const struct expr *expr, struct output_ctx *octx)
 {
 	unsigned int len = div_round_up(expr->len, BITS_PER_BYTE);
@@ -636,8 +616,7 @@ static void inet_protocol_type_print(const struct expr *expr,
 {
 	struct protoent *p;
 
-	if (!nft_output_numeric_proto(octx) &&
-	    mpz_cmp_ui(expr->value, UINT8_MAX) <= 0) {
+	if (!nft_output_numeric_proto(octx)) {
 		p = getprotobynumber(mpz_get_uint8(expr->value));
 		if (p != NULL) {
 			nft_print(octx, "%s", p->p_name);
@@ -716,8 +695,7 @@ static void inet_service_print(const struct expr *expr, struct output_ctx *octx)
 
 void inet_service_type_print(const struct expr *expr, struct output_ctx *octx)
 {
-	if (nft_output_service(octx) &&
-	    mpz_cmp_ui(expr->value, UINT16_MAX) <= 0) {
+	if (nft_output_service(octx)) {
 		inet_service_print(expr, octx);
 		return;
 	}
@@ -861,9 +839,9 @@ const struct datatype mark_type = {
 	.print		= mark_type_print,
 	.json		= mark_type_json,
 	.parse		= mark_type_parse,
+	.flags		= DTYPE_F_PREFIX,
 };
 
-/* symbol table for private datatypes for reject statement. */
 static const struct symbol_table icmp_code_tbl = {
 	.base		= BASE_DECIMAL,
 	.symbols	= {
@@ -879,17 +857,16 @@ static const struct symbol_table icmp_code_tbl = {
 	},
 };
 
-/* private datatype for reject statement. */
-const struct datatype reject_icmp_code_type = {
+const struct datatype icmp_code_type = {
+	.type		= TYPE_ICMP_CODE,
 	.name		= "icmp_code",
-	.desc		= "reject icmp code",
+	.desc		= "icmp code",
 	.size		= BITS_PER_BYTE,
 	.byteorder	= BYTEORDER_BIG_ENDIAN,
 	.basetype	= &integer_type,
 	.sym_tbl	= &icmp_code_tbl,
 };
 
-/* symbol table for private datatypes for reject statement. */
 static const struct symbol_table icmpv6_code_tbl = {
 	.base		= BASE_DECIMAL,
 	.symbols	= {
@@ -903,17 +880,16 @@ static const struct symbol_table icmpv6_code_tbl = {
 	},
 };
 
-/* private datatype for reject statement. */
-const struct datatype reject_icmpv6_code_type = {
+const struct datatype icmpv6_code_type = {
+	.type		= TYPE_ICMPV6_CODE,
 	.name		= "icmpv6_code",
-	.desc		= "reject icmpv6 code",
+	.desc		= "icmpv6 code",
 	.size		= BITS_PER_BYTE,
 	.byteorder	= BYTEORDER_BIG_ENDIAN,
 	.basetype	= &integer_type,
 	.sym_tbl	= &icmpv6_code_tbl,
 };
 
-/* symbol table for private datatypes for reject statement. */
 static const struct symbol_table icmpx_code_tbl = {
 	.base		= BASE_DECIMAL,
 	.symbols	= {
@@ -925,60 +901,6 @@ static const struct symbol_table icmpx_code_tbl = {
 	},
 };
 
-/* private datatype for reject statement. */
-const struct datatype reject_icmpx_code_type = {
-	.name		= "icmpx_code",
-	.desc		= "reject icmpx code",
-	.size		= BITS_PER_BYTE,
-	.byteorder	= BYTEORDER_BIG_ENDIAN,
-	.basetype	= &integer_type,
-	.sym_tbl	= &icmpx_code_tbl,
-};
-
-/* Backward compatible parser for the reject statement. */
-static struct error_record *icmp_code_parse(struct parse_ctx *ctx,
-					    const struct expr *sym,
-					    struct expr **res)
-{
-	return symbolic_constant_parse(ctx, sym, &icmp_code_tbl, res);
-}
-
-const struct datatype icmp_code_type = {
-	.type		= TYPE_ICMP_CODE,
-	.name		= "icmp_code",
-	.desc		= "icmp code",
-	.size		= BITS_PER_BYTE,
-	.byteorder	= BYTEORDER_BIG_ENDIAN,
-	.basetype	= &integer_type,
-	.parse		= icmp_code_parse,
-};
-
-/* Backward compatible parser for the reject statement. */
-static struct error_record *icmpv6_code_parse(struct parse_ctx *ctx,
-					      const struct expr *sym,
-					      struct expr **res)
-{
-	return symbolic_constant_parse(ctx, sym, &icmpv6_code_tbl, res);
-}
-
-const struct datatype icmpv6_code_type = {
-	.type		= TYPE_ICMPV6_CODE,
-	.name		= "icmpv6_code",
-	.desc		= "icmpv6 code",
-	.size		= BITS_PER_BYTE,
-	.byteorder	= BYTEORDER_BIG_ENDIAN,
-	.basetype	= &integer_type,
-	.parse		= icmpv6_code_parse,
-};
-
-/* Backward compatible parser for the reject statement. */
-static struct error_record *icmpx_code_parse(struct parse_ctx *ctx,
-					     const struct expr *sym,
-					     struct expr **res)
-{
-	return symbolic_constant_parse(ctx, sym, &icmpx_code_tbl, res);
-}
-
 const struct datatype icmpx_code_type = {
 	.type		= TYPE_ICMPX_CODE,
 	.name		= "icmpx_code",
@@ -986,13 +908,12 @@ const struct datatype icmpx_code_type = {
 	.size		= BITS_PER_BYTE,
 	.byteorder	= BYTEORDER_BIG_ENDIAN,
 	.basetype	= &integer_type,
-	.parse		= icmpx_code_parse,
+	.sym_tbl	= &icmpx_code_tbl,
 };
 
 void time_print(uint64_t ms, struct output_ctx *octx)
 {
 	uint64_t days, hours, minutes, seconds;
-	bool printed = false;
 
 	if (nft_output_seconds(octx)) {
 		nft_print(octx, "%" PRIu64 "s", ms / 1000);
@@ -1011,29 +932,16 @@ void time_print(uint64_t ms, struct output_ctx *octx)
 	seconds = ms / 1000;
 	ms %= 1000;
 
-	if (days > 0) {
+	if (days > 0)
 		nft_print(octx, "%" PRIu64 "d", days);
-		printed = true;
-	}
-	if (hours > 0) {
+	if (hours > 0)
 		nft_print(octx, "%" PRIu64 "h", hours);
-		printed = true;
-	}
-	if (minutes > 0) {
+	if (minutes > 0)
 		nft_print(octx, "%" PRIu64 "m", minutes);
-		printed = true;
-	}
-	if (seconds > 0) {
+	if (seconds > 0)
 		nft_print(octx, "%" PRIu64 "s", seconds);
-		printed = true;
-	}
-	if (ms > 0) {
+	if (ms > 0)
 		nft_print(octx, "%" PRIu64 "ms", ms);
-		printed = true;
-	}
-
-	if (!printed)
-		nft_print(octx, "0s");
 }
 
 enum {
@@ -1191,7 +1099,6 @@ static struct datatype *dtype_alloc(void)
 
 	dtype = xzalloc(sizeof(*dtype));
 	dtype->flags = DTYPE_F_ALLOC;
-	dtype->refcnt = 1;
 
 	return dtype;
 }
@@ -1209,19 +1116,12 @@ struct datatype *datatype_get(const struct datatype *ptr)
 	return dtype;
 }
 
-void __datatype_set(struct expr *expr, const struct datatype *dtype)
-{
-	const struct datatype *dtype_free;
-
-	dtype_free = expr->dtype;
-	expr->dtype = dtype;
-	datatype_free(dtype_free);
-}
-
 void datatype_set(struct expr *expr, const struct datatype *dtype)
 {
-	if (dtype != expr->dtype)
-		__datatype_set(expr, datatype_get(dtype));
+	if (dtype == expr->dtype)
+		return;
+	datatype_free(expr->dtype);
+	expr->dtype = datatype_get(dtype);
 }
 
 struct datatype *dtype_clone(const struct datatype *orig_dtype)
@@ -1233,7 +1133,7 @@ struct datatype *dtype_clone(const struct datatype *orig_dtype)
 	dtype->name = xstrdup(orig_dtype->name);
 	dtype->desc = xstrdup(orig_dtype->desc);
 	dtype->flags = DTYPE_F_ALLOC | orig_dtype->flags;
-	dtype->refcnt = 1;
+	dtype->refcnt = 0;
 
 	return dtype;
 }
@@ -1246,9 +1146,6 @@ void datatype_free(const struct datatype *ptr)
 		return;
 	if (!(dtype->flags & DTYPE_F_ALLOC))
 		return;
-
-	assert(dtype->refcnt != 0);
-
 	if (--dtype->refcnt > 0)
 		return;
 
@@ -1301,7 +1198,7 @@ const struct datatype *set_datatype_alloc(const struct datatype *orig_dtype,
 
 	/* Restrict dynamic datatype allocation to generic integer datatype. */
 	if (orig_dtype != &integer_type)
-		return datatype_get(orig_dtype);
+		return orig_dtype;
 
 	dtype = dtype_clone(orig_dtype);
 	dtype->byteorder = byteorder;
@@ -1323,7 +1220,7 @@ static struct error_record *time_unit_parse(const struct location *loc,
 	else if (strcmp(str, "week") == 0)
 		*unit = 1ULL * 60 * 60 * 24 * 7;
 	else
-		return error(loc, "Wrong time format, expecting second, minute, hour, day or week");
+		return error(loc, "Wrong rate format");
 
 	return NULL;
 }
@@ -1331,14 +1228,14 @@ static struct error_record *time_unit_parse(const struct location *loc,
 struct error_record *data_unit_parse(const struct location *loc,
 				     const char *str, uint64_t *rate)
 {
-	if (strcmp(str, "bytes") == 0)
+	if (strncmp(str, "bytes", strlen("bytes")) == 0)
 		*rate = 1ULL;
-	else if (strcmp(str, "kbytes") == 0)
+	else if (strncmp(str, "kbytes", strlen("kbytes")) == 0)
 		*rate = 1024;
-	else if (strcmp(str, "mbytes") == 0)
+	else if (strncmp(str, "mbytes", strlen("mbytes")) == 0)
 		*rate = 1024 * 1024;
 	else
-		return error(loc, "Wrong unit format, expecting bytes, kbytes or mbytes");
+		return error(loc, "Wrong rate format");
 
 	return NULL;
 }
@@ -1346,20 +1243,14 @@ struct error_record *data_unit_parse(const struct location *loc,
 struct error_record *rate_parse(const struct location *loc, const char *str,
 				uint64_t *rate, uint64_t *unit)
 {
-	const char *slash, *rate_str;
 	struct error_record *erec;
+	const char *slash;
 
 	slash = strchr(str, '/');
 	if (!slash)
-		return error(loc, "wrong rate format, expecting {bytes,kbytes,mbytes}/{second,minute,hour,day,week}");
-
-	rate_str = strndup(str, slash - str);
-	if (!rate_str)
-		memory_allocation_error();
-
-	erec = data_unit_parse(loc, rate_str, rate);
-	xfree(rate_str);
+		return error(loc, "wrong rate format");
 
+	erec = data_unit_parse(loc, str, rate);
 	if (erec != NULL)
 		return erec;
 
@@ -1379,35 +1270,11 @@ static const struct symbol_table boolean_tbl = {
 	},
 };
 
-static struct error_record *boolean_type_parse(struct parse_ctx *ctx,
-					       const struct expr *sym,
-					       struct expr **res)
-{
-	struct error_record *erec;
-	int num;
-
-	erec = integer_type_parse(ctx, sym, res);
-	if (erec)
-		return erec;
-
-	if (mpz_cmp_ui((*res)->value, 0))
-		num = 1;
-	else
-		num = 0;
-
-	expr_free(*res);
-
-	*res = constant_expr_alloc(&sym->location, &boolean_type,
-				   BYTEORDER_HOST_ENDIAN, 1, &num);
-	return NULL;
-}
-
 const struct datatype boolean_type = {
 	.type		= TYPE_BOOLEAN,
 	.name		= "boolean",
 	.desc		= "boolean type",
 	.size		= 1,
-	.parse		= boolean_type_parse,
 	.basetype	= &integer_type,
 	.sym_tbl	= &boolean_tbl,
 	.json		= boolean_type_json,
