@@ -29,6 +29,7 @@
 extern const struct expr_ops ct_expr_ops;
 extern const struct expr_ops fib_expr_ops;
 extern const struct expr_ops hash_expr_ops;
+extern const struct expr_ops inner_expr_ops;
 extern const struct expr_ops meta_expr_ops;
 extern const struct expr_ops numgen_expr_ops;
 extern const struct expr_ops osf_expr_ops;
@@ -140,10 +141,8 @@ void expr_describe(const struct expr *expr, struct output_ctx *octx)
 		nft_print(octx, "%s expression, datatype %s (%s)",
 			  expr_name(expr), dtype->name, dtype->desc);
 
-		if (dtype == &invalid_type) {
-			nft_print(octx, "\n");
+		if (dtype == &invalid_type)
 			return;
-		}
 	}
 
 	if (dtype->basetype != NULL) {
@@ -380,84 +379,6 @@ struct expr *variable_expr_alloc(const struct location *loc,
 	return expr;
 }
 
-#define NFTNL_UDATA_CONSTANT_TYPE 0
-#define NFTNL_UDATA_CONSTANT_MAX NFTNL_UDATA_CONSTANT_TYPE
-
-#define CONSTANT_EXPR_NFQUEUE_ID 0
-
-static int constant_expr_build_udata(struct nftnl_udata_buf *udbuf,
-				     const struct expr *expr)
-{
-	uint32_t type;
-
-	if (expr->dtype == &queue_type)
-		type = CONSTANT_EXPR_NFQUEUE_ID;
-	else
-		return -1;
-
-	if (!nftnl_udata_put_u32(udbuf, NFTNL_UDATA_CONSTANT_TYPE, type))
-		return -1;
-
-	return 0;
-}
-
-static int constant_parse_udata(const struct nftnl_udata *attr, void *data)
-{
-	const struct nftnl_udata **ud = data;
-	uint8_t type = nftnl_udata_type(attr);
-	uint8_t len = nftnl_udata_len(attr);
-	uint32_t value;
-
-	switch (type) {
-	case NFTNL_UDATA_CONSTANT_TYPE:
-		if (len != sizeof(uint32_t))
-			return -1;
-
-		value = nftnl_udata_get_u32(attr);
-		switch (value) {
-		case CONSTANT_EXPR_NFQUEUE_ID:
-			break;
-		default:
-			return -1;
-		}
-		break;
-	default:
-		return 0;
-	}
-
-	ud[type] = attr;
-
-	return 0;
-}
-
-static struct expr *constant_expr_parse_udata(const struct nftnl_udata *attr)
-{
-	const struct nftnl_udata *ud[NFTNL_UDATA_CONSTANT_MAX + 1] = {};
-	const struct datatype *dtype = NULL;
-	uint32_t type;
-	int err;
-
-	err = nftnl_udata_parse(nftnl_udata_get(attr), nftnl_udata_len(attr),
-				constant_parse_udata, ud);
-	if (err < 0)
-		return NULL;
-
-	if (!ud[NFTNL_UDATA_CONSTANT_TYPE])
-		return NULL;
-
-	type = nftnl_udata_get_u32(ud[NFTNL_UDATA_CONSTANT_TYPE]);
-	switch (type) {
-	case CONSTANT_EXPR_NFQUEUE_ID:
-		dtype = &queue_type;
-		break;
-	default:
-		break;
-	}
-
-	return constant_expr_alloc(&internal_location, dtype, BYTEORDER_HOST_ENDIAN,
-				   16, NULL);
-}
-
 static void constant_expr_print(const struct expr *expr,
 				 struct output_ctx *octx)
 {
@@ -488,8 +409,6 @@ static const struct expr_ops constant_expr_ops = {
 	.cmp		= constant_expr_cmp,
 	.clone		= constant_expr_clone,
 	.destroy	= constant_expr_destroy,
-	.build_udata	= constant_expr_build_udata,
-	.parse_udata	= constant_expr_parse_udata,
 };
 
 struct expr *constant_expr_alloc(const struct location *loc,
@@ -503,7 +422,7 @@ struct expr *constant_expr_alloc(const struct location *loc,
 	expr->flags = EXPR_F_CONSTANT | EXPR_F_SINGLETON;
 
 	mpz_init2(expr->value, len);
-	if (data != NULL && len)
+	if (data != NULL)
 		mpz_import_data(expr->value, data, byteorder,
 				div_round_up(len, BITS_PER_BYTE));
 
@@ -832,17 +751,6 @@ void relational_expr_pctx_update(struct proto_ctx *ctx,
 				    i->key->etype == EXPR_VALUE)
 					ops->pctx_update(ctx, &expr->location, left, i->key);
 			}
-		} else if (ops == &meta_expr_ops &&
-			   right->etype == EXPR_SET_REF) {
-			const struct expr *key = right->set->key;
-			struct expr *tmp;
-
-			tmp = constant_expr_alloc(&expr->location, key->dtype,
-						  key->byteorder, key->len,
-						  NULL);
-
-			ops->pctx_update(ctx, &expr->location, left, tmp);
-			expr_free(tmp);
 		}
 	}
 }
@@ -1087,7 +995,7 @@ static struct expr *concat_expr_parse_udata(const struct nftnl_udata *attr)
 			goto err_free;
 
 		etype = nftnl_udata_get_u32(nest_ud[NFTNL_UDATA_SET_KEY_CONCAT_SUB_TYPE]);
-		ops = expr_ops_by_type_u32(etype);
+		ops = expr_ops_by_type(etype);
 		if (!ops || !ops->parse_udata)
 			goto err_free;
 
@@ -1105,7 +1013,7 @@ static struct expr *concat_expr_parse_udata(const struct nftnl_udata *attr)
 	if (!dtype)
 		goto err_free;
 
-	__datatype_set(concat_expr, dtype);
+	concat_expr->dtype = datatype_get(dtype);
 	concat_expr->len = len;
 
 	return concat_expr;
@@ -1434,8 +1342,9 @@ static void set_elem_expr_destroy(struct expr *expr)
 		stmt_free(stmt);
 }
 
-static void __set_elem_expr_clone(struct expr *new, const struct expr *expr)
+static void set_elem_expr_clone(struct expr *new, const struct expr *expr)
 {
+	new->key = expr_clone(expr->key);
 	new->expiration = expr->expiration;
 	new->timeout = expr->timeout;
 	if (expr->comment)
@@ -1443,12 +1352,6 @@ static void __set_elem_expr_clone(struct expr *new, const struct expr *expr)
 	init_list_head(&new->stmt_list);
 }
 
-static void set_elem_expr_clone(struct expr *new, const struct expr *expr)
-{
-	new->key = expr_clone(expr->key);
-	__set_elem_expr_clone(new, expr);
-}
-
 static const struct expr_ops set_elem_expr_ops = {
 	.type		= EXPR_SET_ELEM,
 	.name		= "set element",
@@ -1476,17 +1379,11 @@ static void set_elem_catchall_expr_print(const struct expr *expr,
 	nft_print(octx, "*");
 }
 
-static void set_elem_catchall_expr_clone(struct expr *new, const struct expr *expr)
-{
-	__set_elem_expr_clone(new, expr);
-}
-
 static const struct expr_ops set_elem_catchall_expr_ops = {
 	.type		= EXPR_SET_ELEM_CATCHALL,
 	.name		= "catch-all set element",
 	.print		= set_elem_catchall_expr_print,
 	.json		= set_elem_catchall_expr_json,
-	.clone		= set_elem_catchall_expr_clone,
 };
 
 struct expr *set_elem_catchall_expr_alloc(const struct location *loc)
@@ -1601,7 +1498,9 @@ void range_expr_value_high(mpz_t rop, const struct expr *expr)
 static const struct expr_ops *__expr_ops_by_type(enum expr_types etype)
 {
 	switch (etype) {
-	case EXPR_INVALID: break;
+	case EXPR_INVALID:
+		BUG("Invalid expression ops requested");
+		break;
 	case EXPR_VERDICT: return &verdict_expr_ops;
 	case EXPR_SYMBOL: return &symbol_expr_ops;
 	case EXPR_VARIABLE: return &variable_expr_ops;
@@ -1633,23 +1532,20 @@ static const struct expr_ops *__expr_ops_by_type(enum expr_types etype)
 	case EXPR_FLAGCMP: return &flagcmp_expr_ops;
 	}
 
-	return NULL;
+	BUG("Unknown expression type %d\n", etype);
 }
 
 const struct expr_ops *expr_ops(const struct expr *e)
 {
-	const struct expr_ops *ops;
-
-	ops = __expr_ops_by_type(e->etype);
-	if (!ops)
-		BUG("Unknown expression type %d\n", e->etype);
-
-	return ops;
+	return __expr_ops_by_type(e->etype);
 }
 
-const struct expr_ops *expr_ops_by_type_u32(uint32_t value)
+const struct expr_ops *expr_ops_by_type(uint32_t value)
 {
-	if (value > EXPR_MAX)
+	/* value might come from unreliable source, such as "udata"
+	 * annotation of set keys.  Avoid BUG() assertion.
+	 */
+	if (value == EXPR_INVALID || value > EXPR_MAX)
 		return NULL;
 
 	return __expr_ops_by_type(value);
