@@ -15,7 +15,6 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/stat.h>
 
 static int nft_netlink(struct nft_ctx *nft,
 		       struct list_head *cmds, struct list_head *msgs,
@@ -56,13 +55,6 @@ static int nft_netlink(struct nft_ctx *nft,
 
 	ret = mnl_batch_talk(&ctx, &err_list, num_cmds);
 	if (ret < 0) {
-		if (ctx.maybe_emsgsize && errno == EMSGSIZE) {
-			netlink_io_error(&ctx, NULL,
-					 "Could not process rule: %s\n"
-					 "Please, rise /proc/sys/net/core/wmem_max on the host namespace. Hint: %d bytes",
-					 strerror(errno), round_pow_2(ctx.maybe_emsgsize));
-			goto out;
-		}
 		netlink_io_error(&ctx, NULL,
 				 "Could not process rule: %s", strerror(errno));
 		goto out;
@@ -159,7 +151,6 @@ void nft_ctx_clear_vars(struct nft_ctx *ctx)
 	}
 	ctx->num_vars = 0;
 	xfree(ctx->vars);
-	ctx->vars = NULL;
 }
 
 EXPORT_SYMBOL(nft_ctx_add_include_path);
@@ -529,8 +520,7 @@ static int nft_evaluate(struct nft_ctx *nft, struct list_head *msgs,
 		collapsed = true;
 
 	list_for_each_entry(cmd, cmds, list) {
-		if (cmd->op != CMD_ADD &&
-		    cmd->op != CMD_CREATE)
+		if (cmd->op != CMD_ADD)
 			continue;
 
 		nft_cmd_expand(cmd);
@@ -555,6 +545,13 @@ static int nft_evaluate(struct nft_ctx *nft, struct list_head *msgs,
 	if (err < 0 || nft->state->nerrs)
 		return -1;
 
+	list_for_each_entry(cmd, cmds, list) {
+		if (cmd->op != CMD_ADD)
+			continue;
+
+		nft_cmd_post_expand(cmd);
+	}
+
 	return 0;
 }
 
@@ -611,10 +608,8 @@ err:
 	    nft_output_json(&nft->output) &&
 	    nft_output_echo(&nft->output))
 		json_print_echo(nft);
-
-	if (rc || nft->check)
+	if (rc)
 		nft_cache_release(&nft->cache);
-
 	return rc;
 }
 
@@ -659,47 +654,13 @@ retry:
 	return rc;
 }
 
-/* need to use stat() to, fopen() will block for named fifos and
- * libjansson makes no checks before or after open either.
- * /dev/stdin is *never* used, read() from STDIN_FILENO is used instead.
- */
-static struct error_record *filename_is_useable(struct nft_ctx *nft, const char *name)
-{
-	unsigned int type;
-	struct stat sb;
-	int err;
-
-	if (!strcmp(name, "/dev/stdin"))
-		return NULL;
-
-	err = stat(name, &sb);
-	if (err)
-		return error(&internal_location, "Could not open file \"%s\": %s\n",
-			     name, strerror(errno));
-
-	type = sb.st_mode & S_IFMT;
-
-	if (type == S_IFREG || type == S_IFIFO)
-		return NULL;
-
-	return error(&internal_location, "Not a regular file: \"%s\"\n", name);
-}
-
 static int __nft_run_cmd_from_filename(struct nft_ctx *nft, const char *filename)
 {
-	struct error_record *erec;
 	struct cmd *cmd, *next;
 	int rc, parser_rc;
 	LIST_HEAD(msgs);
 	LIST_HEAD(cmds);
 
-	erec = filename_is_useable(nft, filename);
-	if (erec) {
-		erec_print(&nft->output, erec, nft->debug_mask);
-		erec_destroy(erec);
-		return -1;
-	}
-
 	rc = load_cmdline_vars(nft, &msgs);
 	if (rc < 0)
 		goto err;
@@ -753,8 +714,7 @@ err:
 	    nft_output_json(&nft->output) &&
 	    nft_output_echo(&nft->output))
 		json_print_echo(nft);
-
-	if (rc || nft->check)
+	if (rc)
 		nft_cache_release(&nft->cache);
 
 	scope_release(nft->state->scopes[0]);
@@ -792,7 +752,8 @@ int nft_run_cmd_from_filename(struct nft_ctx *nft, const char *filename)
 	if (!strcmp(filename, "-"))
 		filename = "/dev/stdin";
 
-	if (!strcmp(filename, "/dev/stdin"))
+	if (!strcmp(filename, "/dev/stdin") &&
+	    !nft_output_json(&nft->output))
 		nft->stdin_buf = stdin_to_buffer();
 
 	if (nft->optimize_flags) {
