@@ -11,6 +11,8 @@
 #include <gmputil.h>
 #include <list.h>
 
+static void expr_msort_value(const struct expr *expr, mpz_t value);
+
 static void concat_expr_msort_value(const struct expr *expr, mpz_t value)
 {
 	unsigned int len = 0, ilen;
@@ -26,17 +28,20 @@ static void concat_expr_msort_value(const struct expr *expr, mpz_t value)
 	mpz_import_data(value, data, BYTEORDER_HOST_ENDIAN, len);
 }
 
-static mpz_srcptr expr_msort_value(const struct expr *expr, mpz_t value)
+static void expr_msort_value(const struct expr *expr, mpz_t value)
 {
 	switch (expr->etype) {
 	case EXPR_SET_ELEM:
-		return expr_msort_value(expr->key, value);
+		expr_msort_value(expr->key, value);
+		break;
 	case EXPR_BINOP:
 	case EXPR_MAPPING:
 	case EXPR_RANGE:
-		return expr_msort_value(expr->left, value);
+		expr_msort_value(expr->left, value);
+		break;
 	case EXPR_VALUE:
-		return expr->value;
+		mpz_set(value, expr->value);
+		break;
 	case EXPR_CONCAT:
 		concat_expr_msort_value(expr, value);
 		break;
@@ -47,24 +52,20 @@ static mpz_srcptr expr_msort_value(const struct expr *expr, mpz_t value)
 	default:
 		BUG("Unknown expression %s\n", expr_name(expr));
 	}
-	return value;
 }
 
 static int expr_msort_cmp(const struct expr *e1, const struct expr *e2)
 {
-	mpz_srcptr value1;
-	mpz_srcptr value2;
-	mpz_t value1_tmp;
-	mpz_t value2_tmp;
+	mpz_t value1, value2;
 	int ret;
 
-	mpz_init(value1_tmp);
-	mpz_init(value2_tmp);
-	value1 = expr_msort_value(e1, value1_tmp);
-	value2 = expr_msort_value(e2, value2_tmp);
+	mpz_init(value1);
+	mpz_init(value2);
+	expr_msort_value(e1, value1);
+	expr_msort_value(e2, value2);
 	ret = mpz_cmp(value1, value2);
-	mpz_clear(value1_tmp);
-	mpz_clear(value2_tmp);
+	mpz_clear(value1);
+	mpz_clear(value2);
 
 	return ret;
 }
@@ -77,7 +78,7 @@ void list_splice_sorted(struct list_head *list, struct list_head *head)
 	while (l != list) {
 		if (h == head ||
 		    expr_msort_cmp(list_entry(l, typeof(struct expr), list),
-				   list_entry(h, typeof(struct expr), list)) <= 0) {
+				   list_entry(h, typeof(struct expr), list)) < 0) {
 			l = l->next;
 			list_add_tail(l->prev, h);
 			continue;
