@@ -66,39 +66,50 @@ static struct error_record *tchandle_type_parse(struct parse_ctx *ctx,
 						struct expr **res)
 {
 	uint32_t handle;
+	char *str = NULL;
 
 	if (strcmp(sym->identifier, "root") == 0)
 		handle = TC_H_ROOT;
 	else if (strcmp(sym->identifier, "none") == 0)
 		handle = TC_H_UNSPEC;
 	else if (strchr(sym->identifier, ':')) {
-		char *colon, *end;
 		uint32_t tmp;
+		char *colon;
 
-		errno = 0;
-		tmp = strtoul(sym->identifier, &colon, 16);
-		if (errno != 0 || sym->identifier == colon)
+		str = xstrdup(sym->identifier);
+
+		colon = strchr(str, ':');
+		if (!colon)
 			goto err;
 
-		if (*colon != ':')
+		*colon = '\0';
+
+		errno = 0;
+		tmp = strtoull(str, NULL, 16);
+		if (errno != 0)
 			goto err;
 
-		handle = tmp << 16;
+		handle = (tmp << 16);
+		if (str[strlen(str) - 1] == ':')
+			goto out;
+
 		errno = 0;
-		tmp = strtoul(colon + 1, &end, 16);
-		if (errno != 0 || *end)
+		tmp = strtoull(colon + 1, NULL, 16);
+		if (errno != 0)
 			goto err;
 
 		handle |= tmp;
 	} else {
 		handle = strtoull(sym->identifier, NULL, 0);
 	}
-
+out:
+	xfree(str);
 	*res = constant_expr_alloc(&sym->location, sym->dtype,
 				   BYTEORDER_HOST_ENDIAN,
 				   sizeof(handle) * BITS_PER_BYTE, &handle);
 	return NULL;
 err:
+	xfree(str);
 	return error(&sym->location, "Could not parse %s", sym->dtype->desc);
 }
 
@@ -360,6 +371,7 @@ const struct datatype devgroup_type = {
 	.print		= devgroup_type_print,
 	.json		= devgroup_type_json,
 	.parse		= devgroup_type_parse,
+	.flags		= DTYPE_F_PREFIX,
 };
 
 const struct datatype ifname_type = {
@@ -479,21 +491,16 @@ static void day_type_print(const struct expr *expr, struct output_ctx *octx)
 static void hour_type_print(const struct expr *expr, struct output_ctx *octx)
 {
 	uint32_t seconds = mpz_get_uint32(expr->value), minutes, hours;
-	struct tm cur_tm;
+	struct tm *cur_tm;
 	time_t ts;
 
 	/* Obtain current tm, so that we can add tm_gmtoff */
 	ts = time(NULL);
-	if (ts != ((time_t) -1) && localtime_r(&ts, &cur_tm)) {
-		int32_t adj = seconds + cur_tm.tm_gmtoff;
+	cur_tm = localtime(&ts);
 
-		if (adj < 0)
-			adj += SECONDS_PER_DAY;
-		else if (adj >= SECONDS_PER_DAY)
-			adj -= SECONDS_PER_DAY;
+	if (cur_tm)
+		seconds = (seconds + cur_tm->tm_gmtoff) % SECONDS_PER_DAY;
 
-		seconds = adj;
-	}
 	minutes = seconds / 60;
 	seconds %= 60;
 	hours = minutes / 60;
@@ -710,16 +717,12 @@ static bool meta_key_is_unqualified(enum nft_meta_keys key)
 
 static void meta_expr_print(const struct expr *expr, struct output_ctx *octx)
 {
-	const char *token = "unknown";
-	uint32_t key = expr->meta.key;
-
-	if (key < array_size(meta_templates))
-		token = meta_templates[key].token;
-
-	if (meta_key_is_unqualified(key))
-		nft_print(octx, "%s", token);
+	if (meta_key_is_unqualified(expr->meta.key))
+		nft_print(octx, "%s",
+			  meta_templates[expr->meta.key].token);
 	else
-		nft_print(octx, "meta %s", token);
+		nft_print(octx, "meta %s",
+			  meta_templates[expr->meta.key].token);
 }
 
 static bool meta_expr_cmp(const struct expr *e1, const struct expr *e2)
@@ -731,6 +734,7 @@ static void meta_expr_clone(struct expr *new, const struct expr *expr)
 {
 	new->meta.key = expr->meta.key;
 	new->meta.base = expr->meta.base;
+	new->meta.inner_desc = expr->meta.inner_desc;
 }
 
 /**
@@ -804,13 +808,19 @@ static void meta_expr_pctx_update(struct proto_ctx *ctx,
 }
 
 #define NFTNL_UDATA_META_KEY 0
-#define NFTNL_UDATA_META_MAX 1
+#define NFTNL_UDATA_META_INNER_DESC 1
+#define NFTNL_UDATA_META_MAX 2
 
 static int meta_expr_build_udata(struct nftnl_udata_buf *udbuf,
 				 const struct expr *expr)
 {
 	nftnl_udata_put_u32(udbuf, NFTNL_UDATA_META_KEY, expr->meta.key);
 
+	if (expr->meta.inner_desc) {
+		nftnl_udata_put_u32(udbuf, NFTNL_UDATA_META_INNER_DESC,
+				    expr->meta.inner_desc->id);
+	}
+
 	return 0;
 }
 
@@ -822,6 +832,7 @@ static int meta_parse_udata(const struct nftnl_udata *attr, void *data)
 
 	switch (type) {
 	case NFTNL_UDATA_META_KEY:
+	case NFTNL_UDATA_META_INNER_DESC:
 		if (len != sizeof(uint32_t))
 			return -1;
 		break;
@@ -836,6 +847,8 @@ static int meta_parse_udata(const struct nftnl_udata *attr, void *data)
 static struct expr *meta_expr_parse_udata(const struct nftnl_udata *attr)
 {
 	const struct nftnl_udata *ud[NFTNL_UDATA_META_MAX + 1] = {};
+	const struct proto_desc *desc;
+	struct expr *expr;
 	uint32_t key;
 	int err;
 
@@ -849,7 +862,14 @@ static struct expr *meta_expr_parse_udata(const struct nftnl_udata *attr)
 
 	key = nftnl_udata_get_u32(ud[NFTNL_UDATA_META_KEY]);
 
-	return meta_expr_alloc(&internal_location, key);
+	expr = meta_expr_alloc(&internal_location, key);
+
+	if (ud[NFTNL_UDATA_META_INNER_DESC]) {
+		desc = find_proto_desc(ud[NFTNL_UDATA_META_INNER_DESC]);
+		expr->meta.inner_desc = desc;
+	}
+
+	return expr;
 }
 
 const struct expr_ops meta_expr_ops = {
@@ -898,16 +918,12 @@ struct expr *meta_expr_alloc(const struct location *loc, enum nft_meta_keys key)
 
 static void meta_stmt_print(const struct stmt *stmt, struct output_ctx *octx)
 {
-	const char *token = "unknown";
-	uint32_t key = stmt->meta.key;
-
-	if (key < array_size(meta_templates))
-		token = meta_templates[key].token;
-
 	if (meta_key_is_unqualified(stmt->meta.key))
-		nft_print(octx, "%s set ", token);
+		nft_print(octx, "%s set ",
+			  meta_templates[stmt->meta.key].token);
 	else
-		nft_print(octx, "meta %s set ", token);
+		nft_print(octx, "meta %s set ",
+			  meta_templates[stmt->meta.key].token);
 
 	expr_print(stmt->meta.expr, octx);
 }
@@ -932,11 +948,8 @@ struct stmt *meta_stmt_alloc(const struct location *loc, enum nft_meta_keys key,
 
 	stmt = stmt_alloc(loc, &meta_stmt_ops);
 	stmt->meta.key	= key;
+	stmt->meta.tmpl	= &meta_templates[key];
 	stmt->meta.expr	= expr;
-
-        if (key < array_size(meta_templates))
-                stmt->meta.tmpl = &meta_templates[key];
-
 	return stmt;
 }
 
