@@ -127,19 +127,6 @@ struct nftnl_obj *netlink_obj_alloc(const struct nlmsghdr *nlh)
 	return nlo;
 }
 
-struct nftnl_flowtable *netlink_flowtable_alloc(const struct nlmsghdr *nlh)
-{
-	struct nftnl_flowtable *nlf;
-
-	nlf = nftnl_flowtable_alloc();
-	if (nlf == NULL)
-		memory_allocation_error();
-	if (nftnl_flowtable_nlmsg_parse(nlh, nlf) < 0)
-		netlink_abi_error();
-
-	return nlf;
-}
-
 static uint32_t netlink_msg2nftnl_of(uint32_t type, uint16_t flags)
 {
 	switch (type) {
@@ -237,10 +224,6 @@ static int netlink_events_table_cb(const struct nlmsghdr *nlh, int type,
 
 	nlt = netlink_table_alloc(nlh);
 	t = netlink_delinearize_table(monh->ctx, nlt);
-	if (!t) {
-		nftnl_table_free(nlt);
-		return MNL_CB_ERROR;
-	}
 	cmd = netlink_msg2cmd(type, nlh->nlmsg_flags);
 
 	switch (monh->format) {
@@ -404,19 +387,13 @@ static bool netlink_event_range_cache(struct set *cached_set,
 
 	/* don't cache half-open range elements */
 	elem = list_entry(dummyset->init->expressions.prev, struct expr, list);
-	if (!set_elem_is_open_interval(elem) &&
-	    dummyset->desc.field_count <= 1) {
+	if (!set_elem_is_open_interval(elem)) {
 		cached_set->rg_cache = expr_clone(elem);
 		return true;
 	}
 
 out_decompose:
-	if (dummyset->flags & NFT_SET_INTERVAL &&
-	    dummyset->desc.field_count > 1)
-		concat_range_aggregate(dummyset->init);
-	else
-		interval_map_decompose(dummyset->init);
-
+	interval_map_decompose(dummyset->init);
 	return false;
 }
 
@@ -457,7 +434,6 @@ static int netlink_events_setelem_cb(const struct nlmsghdr *nlh, int type,
 		dummyset->data = expr_clone(set->data);
 	dummyset->flags = set->flags;
 	dummyset->init = set_expr_alloc(monh->loc, set);
-	dummyset->desc.field_count = set->desc.field_count;
 
 	nlsei = nftnl_set_elems_iter_create(nls);
 	if (nlsei == NULL)
@@ -470,8 +446,8 @@ static int netlink_events_setelem_cb(const struct nlmsghdr *nlh, int type,
 			nftnl_set_elems_iter_destroy(nlsei);
 			goto out;
 		}
-		if (netlink_delinearize_setelem(monh->ctx,
-						nlse, dummyset) < 0) {
+		if (netlink_delinearize_setelem(nlse, dummyset,
+						&monh->ctx->nft->cache) < 0) {
 			set_free(dummyset);
 			nftnl_set_elems_iter_destroy(nlsei);
 			goto out;
@@ -556,50 +532,6 @@ static int netlink_events_obj_cb(const struct nlmsghdr *nlh, int type,
 	return MNL_CB_OK;
 }
 
-static int netlink_events_flowtable_cb(const struct nlmsghdr *nlh, int type,
-				       struct netlink_mon_handler *monh)
-{
-	const char *family, *cmd;
-	struct nftnl_flowtable *nlf;
-	struct flowtable *ft;
-
-	nlf = netlink_flowtable_alloc(nlh);
-
-	ft = netlink_delinearize_flowtable(monh->ctx, nlf);
-	if (!ft) {
-		nftnl_flowtable_free(nlf);
-		return MNL_CB_ERROR;
-	}
-	family = family2str(ft->handle.family);
-	cmd = netlink_msg2cmd(type, nlh->nlmsg_flags);
-
-	switch (monh->format) {
-	case NFTNL_OUTPUT_DEFAULT:
-		nft_mon_print(monh, "%s ", cmd);
-
-		switch (type) {
-		case NFT_MSG_NEWFLOWTABLE:
-			flowtable_print_plain(ft, &monh->ctx->nft->output);
-			break;
-		case NFT_MSG_DELFLOWTABLE:
-			nft_mon_print(monh, "flowtable %s %s %s", family,
-				      ft->handle.table.name,
-				      ft->handle.flowtable.name);
-			break;
-		}
-		nft_mon_print(monh, "\n");
-		break;
-	case NFTNL_OUTPUT_JSON:
-		monitor_print_flowtable_json(monh, cmd, ft);
-		if (!nft_output_echo(&monh->ctx->nft->output))
-			nft_mon_print(monh, "\n");
-		break;
-	}
-	flowtable_free(ft);
-	nftnl_flowtable_free(nlf);
-	return MNL_CB_OK;
-}
-
 static void rule_map_decompose_cb(struct set *s, void *data)
 {
 	if (!set_is_anonymous(s->flags))
@@ -772,7 +704,8 @@ static void netlink_events_cache_addsetelem(struct netlink_mon_handler *monh,
 
 	nlse = nftnl_set_elems_iter_next(nlsei);
 	while (nlse != NULL) {
-		if (netlink_delinearize_setelem(monh->ctx, nlse, set) < 0) {
+		if (netlink_delinearize_setelem(nlse, set,
+						&monh->ctx->nft->cache) < 0) {
 			fprintf(stderr,
 				"W: Unable to cache set_elem. "
 				"Delinearize failed.\n");
@@ -1019,10 +952,6 @@ static int netlink_events_cb(const struct nlmsghdr *nlh, void *data)
 	case NFT_MSG_DELOBJ:
 		ret = netlink_events_obj_cb(nlh, type, monh);
 		break;
-	case NFT_MSG_NEWFLOWTABLE:
-	case NFT_MSG_DELFLOWTABLE:
-		ret = netlink_events_flowtable_cb(nlh, type, monh);
-		break;
 	case NFT_MSG_NEWGEN:
 		ret = netlink_events_newgen_cb(nlh, type, monh);
 		break;
