@@ -254,17 +254,10 @@ static int netlink_export_pad(unsigned char *data, const mpz_t v,
 	return netlink_padded_len(i->len) / BITS_PER_BYTE;
 }
 
-static void byteorder_switch_expr_value(mpz_t v, const struct expr *e)
-{
-	mpz_switch_byteorder(v, div_round_up(e->len, BITS_PER_BYTE));
-}
-
 static int __netlink_gen_concat_key(uint32_t flags, const struct expr *i,
 				    unsigned char *data)
 {
 	struct expr *expr;
-	mpz_t value;
-	int ret;
 
 	switch (i->etype) {
 	case EXPR_RANGE:
@@ -273,11 +266,9 @@ static int __netlink_gen_concat_key(uint32_t flags, const struct expr *i,
 		else
 			expr = i->left;
 
-		mpz_init_set(value, expr->value);
-
 		if (expr_basetype(expr)->type == TYPE_INTEGER &&
 		    expr->byteorder == BYTEORDER_HOST_ENDIAN)
-			byteorder_switch_expr_value(value, expr);
+			mpz_switch_byteorder(expr->value, expr->len / BITS_PER_BYTE);
 
 		i = expr;
 		break;
@@ -289,7 +280,7 @@ static int __netlink_gen_concat_key(uint32_t flags, const struct expr *i,
 			mpz_init_bitmask(v, i->len - i->prefix_len);
 
 			if (i->byteorder == BYTEORDER_HOST_ENDIAN)
-				byteorder_switch_expr_value(v, i);
+				mpz_switch_byteorder(v, i->len / BITS_PER_BYTE);
 
 			mpz_add(v, i->prefix->value, v);
 			count = netlink_export_pad(data, v, i);
@@ -298,56 +289,38 @@ static int __netlink_gen_concat_key(uint32_t flags, const struct expr *i,
 		}
 		return netlink_export_pad(data, i->prefix->value, i);
 	case EXPR_VALUE:
-		mpz_init_set(value, i->value);
-
-		/* Switch byteorder to big endian representation when the set
+		/* Switch byteorder only once for singleton values when the set
 		 * contains concatenation of intervals.
 		 */
-		if (!(flags & (EXPR_F_INTERVAL| EXPR_F_INTERVAL_END)))
+		if (!(flags & EXPR_F_INTERVAL))
 			break;
 
 		expr = (struct expr *)i;
 		if (expr_basetype(expr)->type == TYPE_INTEGER &&
 		    expr->byteorder == BYTEORDER_HOST_ENDIAN)
-			byteorder_switch_expr_value(value, expr);
+			mpz_switch_byteorder(expr->value, expr->len / BITS_PER_BYTE);
 		break;
 	default:
 		BUG("invalid expression type '%s' in set", expr_ops(i)->name);
 	}
 
-	ret = netlink_export_pad(data, value, i);
-	mpz_clear(value);
-
-	return ret;
-}
-
-static void nft_data_memcpy(struct nft_data_linearize *nld,
-			    const void *src, unsigned int len)
-{
-	if (len > sizeof(nld->value))
-		BUG("nld buffer overflow: want to copy %u, max %u\n", len, (unsigned int)sizeof(nld->value));
-
-	memcpy(nld->value, src, len);
-	nld->len = len;
+	return netlink_export_pad(data, i->value, i);
 }
 
 static void netlink_gen_concat_key(const struct expr *expr,
 				    struct nft_data_linearize *nld)
 {
-	unsigned int len = netlink_padded_len(expr->len) / BITS_PER_BYTE;
-	unsigned char data[NFT_MAX_EXPR_LEN_BYTES];
-	unsigned int offset = 0;
+	unsigned int len = expr->len / BITS_PER_BYTE, offset = 0;
+	unsigned char data[len];
 	const struct expr *i;
 
-	if (len > sizeof(data))
-		BUG("Value export of %u bytes would overflow", len);
-
-	memset(data, 0, sizeof(data));
+	memset(data, 0, len);
 
 	list_for_each_entry(i, &expr->expressions, list)
 		offset += __netlink_gen_concat_key(expr->flags, i, data + offset);
 
-	nft_data_memcpy(nld, data, len);
+	memcpy(nld->value, data, len);
+	nld->len = len;
 }
 
 static int __netlink_gen_concat_data(int end, const struct expr *i,
@@ -381,15 +354,11 @@ static int __netlink_gen_concat_data(int end, const struct expr *i,
 static void __netlink_gen_concat_expand(const struct expr *expr,
 				        struct nft_data_linearize *nld)
 {
-	unsigned int len = (netlink_padded_len(expr->len) / BITS_PER_BYTE) * 2;
-	unsigned char data[NFT_MAX_EXPR_LEN_BYTES];
-	unsigned int offset = 0;
+	unsigned int len = div_round_up(expr->len, BITS_PER_BYTE) * 2, offset = 0;
+	unsigned char data[len];
 	const struct expr *i;
 
-	if (len > sizeof(data))
-		BUG("Value export of %u bytes would overflow", len);
-
-	memset(data, 0, sizeof(data));
+	memset(data, 0, len);
 
 	list_for_each_entry(i, &expr->expressions, list)
 		offset += __netlink_gen_concat_data(false, i, data + offset);
@@ -397,26 +366,24 @@ static void __netlink_gen_concat_expand(const struct expr *expr,
 	list_for_each_entry(i, &expr->expressions, list)
 		offset += __netlink_gen_concat_data(true, i, data + offset);
 
-	nft_data_memcpy(nld, data, len);
+	memcpy(nld->value, data, len);
+	nld->len = len;
 }
 
 static void __netlink_gen_concat(const struct expr *expr,
 				 struct nft_data_linearize *nld)
 {
-	unsigned int len = netlink_padded_len(expr->len) / BITS_PER_BYTE;
-	unsigned char data[NFT_MAX_EXPR_LEN_BYTES];
-	unsigned int offset = 0;
+	unsigned int len = expr->len / BITS_PER_BYTE, offset = 0;
+	unsigned char data[len];
 	const struct expr *i;
 
-	if (len > sizeof(data))
-		BUG("Value export of %u bytes would overflow", len);
-
-	memset(data, 0, sizeof(data));
+	memset(data, 0, len);
 
 	list_for_each_entry(i, &expr->expressions, list)
 		offset += __netlink_gen_concat_data(expr->flags, i, data + offset);
 
-	nft_data_memcpy(nld, data, len);
+	memcpy(nld->value, data, len);
+	nld->len = len;
 }
 
 static void netlink_gen_concat_data(const struct expr *expr,
@@ -478,37 +445,33 @@ static void netlink_gen_verdict(const struct expr *expr,
 static void netlink_gen_range(const struct expr *expr,
 			      struct nft_data_linearize *nld)
 {
-	unsigned int len = (netlink_padded_len(expr->left->len) / BITS_PER_BYTE) * 2;
-	unsigned char data[NFT_MAX_EXPR_LEN_BYTES];
-	unsigned int offset;
+	unsigned int len = div_round_up(expr->left->len, BITS_PER_BYTE) * 2;
+	unsigned char data[len];
+	unsigned int offset = 0;
 
-	if (len > sizeof(data))
-		BUG("Value export of %u bytes would overflow", len);
-
-	memset(data, 0, sizeof(data));
+	memset(data, 0, len);
 	offset = netlink_export_pad(data, expr->left->value, expr->left);
 	netlink_export_pad(data + offset, expr->right->value, expr->right);
-	nft_data_memcpy(nld, data, len);
+	memcpy(nld->value, data, len);
+	nld->len = len;
 }
 
 static void netlink_gen_prefix(const struct expr *expr,
 			       struct nft_data_linearize *nld)
 {
-	unsigned int len = (netlink_padded_len(expr->len) / BITS_PER_BYTE) * 2;
-	unsigned char data[NFT_MAX_EXPR_LEN_BYTES];
+	unsigned int len = div_round_up(expr->len, BITS_PER_BYTE) * 2;
+	unsigned char data[len];
 	int offset;
 	mpz_t v;
 
-	if (len > sizeof(data))
-		BUG("Value export of %u bytes would overflow", len);
-
 	offset = netlink_export_pad(data, expr->prefix->value, expr);
 	mpz_init_bitmask(v, expr->len - expr->prefix_len);
 	mpz_add(v, expr->prefix->value, v);
 	netlink_export_pad(data + offset, v, expr->prefix);
 	mpz_clear(v);
 
-	nft_data_memcpy(nld, data, len);
+	memcpy(nld->value, data, len);
+	nld->len = len;
 }
 
 static void netlink_gen_key(const struct expr *expr,
@@ -654,7 +617,7 @@ static int qsort_device_cmp(const void *a, const void *b)
 struct chain *netlink_delinearize_chain(struct netlink_ctx *ctx,
 					const struct nftnl_chain *nlc)
 {
-	const struct nftnl_udata *ud[NFTNL_UDATA_CHAIN_MAX + 1] = {};
+	const struct nftnl_udata *ud[NFTNL_UDATA_OBJ_MAX + 1] = {};
 	int priority, policy, len = 0, i;
 	const char * const *dev_array;
 	struct chain *chain;
@@ -789,8 +752,7 @@ static int list_table_cb(struct nftnl_table *nlt, void *arg)
 	struct table *table;
 
 	table = netlink_delinearize_table(ctx, nlt);
-	if (table)
-		list_add_tail(&table->list, &ctx->list);
+	list_add_tail(&table->list, &ctx->list);
 
 	return 0;
 }
@@ -833,10 +795,6 @@ enum nft_data_types dtype_map_to_kernel(const struct datatype *dtype)
 
 static const struct datatype *dtype_map_from_kernel(enum nft_data_types type)
 {
-	/* The function always returns ownership of a reference. But for
-	 * &verdict_Type and datatype_lookup(), those are static instances,
-	 * we can omit the datatype_get() call.
-	 */
 	switch (type) {
 	case NFT_DATA_VERDICT:
 		return &verdict_type;
@@ -913,8 +871,8 @@ static struct expr *set_make_key(const struct nftnl_udata *attr)
 {
 	const struct nftnl_udata *ud[NFTNL_UDATA_SET_TYPEOF_MAX + 1] = {};
 	const struct expr_ops *ops;
+	enum expr_types etype;
 	struct expr *expr;
-	uint32_t etype;
 	int err;
 
 	if (!attr)
@@ -930,9 +888,7 @@ static struct expr *set_make_key(const struct nftnl_udata *attr)
 		return NULL;
 
 	etype = nftnl_udata_get_u32(ud[NFTNL_UDATA_SET_TYPEOF_EXPR]);
-	ops = expr_ops_by_type_u32(etype);
-	if (!ops || !ops->parse_udata)
-		return NULL;
+	ops = expr_ops_by_type(etype);
 
 	expr = ops->parse_udata(ud[NFTNL_UDATA_SET_TYPEOF_DATA]);
 	if (!expr)
@@ -963,8 +919,7 @@ static int set_elem_parse_expressions(struct nftnl_expr *e, void *data)
 	struct stmt *stmt;
 
 	stmt = netlink_parse_set_expr(set, cache, e);
-	if (stmt)
-		list_add_tail(&stmt->list, &setelem_parse_ctx->stmt_list);
+	list_add_tail(&stmt->list, &setelem_parse_ctx->stmt_list);
 
 	return 0;
 }
@@ -975,14 +930,12 @@ struct set *netlink_delinearize_set(struct netlink_ctx *ctx,
 	const struct nftnl_udata *ud[NFTNL_UDATA_SET_MAX + 1] = {};
 	enum byteorder keybyteorder = BYTEORDER_INVALID;
 	enum byteorder databyteorder = BYTEORDER_INVALID;
+	const struct datatype *keytype, *datatype = NULL;
 	struct expr *typeof_expr_key, *typeof_expr_data;
 	struct setelem_parse_ctx set_parse_ctx;
-	const struct datatype *datatype = NULL;
-	const struct datatype *keytype = NULL;
-	const struct datatype *dtype2 = NULL;
-	const struct datatype *dtype = NULL;
 	const char *udata, *comment = NULL;
 	uint32_t flags, key, objtype = 0;
+	const struct datatype *dtype;
 	uint32_t data_interval = 0;
 	bool automerge = false;
 	struct set *set;
@@ -1034,8 +987,8 @@ struct set *netlink_delinearize_set(struct netlink_ctx *ctx,
 			netlink_io_error(ctx, NULL,
 					 "Unknown data type in set key %u",
 					 data);
-			set = NULL;
-			goto out;
+			datatype_free(keytype);
+			return NULL;
 		}
 	}
 
@@ -1070,29 +1023,17 @@ struct set *netlink_delinearize_set(struct netlink_ctx *ctx,
 	}
 	list_splice_tail(&set_parse_ctx.stmt_list, &set->stmt_list);
 
-	set->flags = nftnl_set_get_u32(nls, NFTNL_SET_FLAGS);
-
 	if (datatype) {
-		uint32_t dlen;
-
-		dtype2 = set_datatype_alloc(datatype, databyteorder);
+		dtype = set_datatype_alloc(datatype, databyteorder);
 		klen = nftnl_set_get_u32(nls, NFTNL_SET_DATA_LEN) * BITS_PER_BYTE;
 
-		dlen = data_interval ?  klen / 2 : klen;
-
-		if (set_udata_key_valid(typeof_expr_data, dlen)) {
-			typeof_expr_data->len = klen;
+		if (set_udata_key_valid(typeof_expr_data, klen)) {
+			datatype_free(datatype_get(dtype));
 			set->data = typeof_expr_data;
-			typeof_expr_data = NULL;
-		} else if (set->flags & NFT_SET_OBJECT) {
-			set->data = constant_expr_alloc(&netlink_location,
-							dtype2,
-							databyteorder, klen,
-							NULL);
 		} else {
 			expr_free(typeof_expr_data);
 			set->data = constant_expr_alloc(&netlink_location,
-							dtype2,
+							dtype,
 							databyteorder, klen,
 							NULL);
 
@@ -1103,12 +1044,16 @@ struct set *netlink_delinearize_set(struct netlink_ctx *ctx,
 
 		if (data_interval)
 			set->data->flags |= EXPR_F_INTERVAL;
+
+		if (dtype != datatype)
+			datatype_free(datatype);
 	}
 
 	dtype = set_datatype_alloc(keytype, keybyteorder);
 	klen = nftnl_set_get_u32(nls, NFTNL_SET_KEY_LEN) * BITS_PER_BYTE;
 
 	if (set_udata_key_valid(typeof_expr_key, klen)) {
+		datatype_free(datatype_get(dtype));
 		set->key = typeof_expr_key;
 		set->key_typeof_valid = true;
 	} else {
@@ -1118,6 +1063,10 @@ struct set *netlink_delinearize_set(struct netlink_ctx *ctx,
 					       NULL);
 	}
 
+	if (dtype != keytype)
+		datatype_free(keytype);
+
+	set->flags   = nftnl_set_get_u32(nls, NFTNL_SET_FLAGS);
 	set->handle.handle.id = nftnl_set_get_u64(nls, NFTNL_SET_HANDLE);
 
 	set->objtype = objtype;
@@ -1144,11 +1093,6 @@ struct set *netlink_delinearize_set(struct netlink_ctx *ctx,
 		}
 	}
 
-out:
-	datatype_free(datatype);
-	datatype_free(keytype);
-	datatype_free(dtype2);
-	datatype_free(dtype);
 	return set;
 }
 
@@ -1235,15 +1179,10 @@ static struct expr *range_expr_reduce(struct expr *range)
 static struct expr *netlink_parse_interval_elem(const struct set *set,
 						struct expr *expr)
 {
-	unsigned int len = netlink_padded_len(expr->len) / BITS_PER_BYTE;
+	unsigned int len = div_round_up(expr->len, BITS_PER_BYTE);
 	const struct datatype *dtype = set->data->dtype;
 	struct expr *range, *left, *right;
-	char data[NFT_MAX_EXPR_LEN_BYTES];
-
-	if (len > sizeof(data))
-		BUG("Value export of %u bytes would overflow", len);
-
-	memset(data, 0, sizeof(data));
+	char data[len];
 
 	mpz_export_data(data, expr->value, dtype->byteorder, len);
 	left = constant_expr_alloc(&internal_location, dtype,
@@ -1402,13 +1341,12 @@ static void set_elem_parse_udata(struct nftnl_set_elem *nlse,
 			nftnl_udata_get_u32(ud[NFTNL_UDATA_SET_ELEM_FLAGS]);
 }
 
-int netlink_delinearize_setelem(struct netlink_ctx *ctx,
-				struct nftnl_set_elem *nlse,
-				struct set *set)
+int netlink_delinearize_setelem(struct nftnl_set_elem *nlse,
+				struct set *set, struct nft_cache *cache)
 {
 	struct setelem_parse_ctx setelem_parse_ctx = {
 		.set	= set,
-		.cache	= &ctx->nft->cache,
+		.cache	= cache,
 	};
 	struct nft_data_delinearize nld;
 	struct expr *expr, *key, *data;
@@ -1442,9 +1380,7 @@ key_end:
 		key->byteorder = set->key->byteorder;
 		key->len = set->key->len;
 	} else {
-		netlink_io_error(ctx, NULL,
-			         "Unexpected set element with no key");
-		return 0;
+		BUG("Unexpected set element with no key\n");
 	}
 
 	expr = set_elem_expr_alloc(&netlink_location, key);
@@ -1461,7 +1397,7 @@ key_end:
 		struct stmt *stmt;
 
 		nle = nftnl_set_elem_get(nlse, NFTNL_SET_ELEM_EXPR, NULL);
-		stmt = netlink_parse_set_expr(set, &ctx->nft->cache, nle);
+		stmt = netlink_parse_set_expr(set, cache, nle);
 		list_add_tail(&stmt->list, &setelem_parse_ctx.stmt_list);
 	} else if (nftnl_set_elem_is_set(nlse, NFTNL_SET_ELEM_EXPRESSIONS)) {
 		nftnl_set_elem_expr_foreach(nlse, set_elem_parse_expressions,
@@ -1490,11 +1426,7 @@ key_end:
 		data = netlink_alloc_data(&netlink_location, &nld,
 					  set->data->dtype->type == TYPE_VERDICT ?
 					  NFT_REG_VERDICT : NFT_REG_1);
-
-		if (set->data->dtype->is_typeof)
-			datatype_set(data, set->data->dtype->basetype);
-		else
-			datatype_set(data, set->data->dtype);
+		datatype_set(data, set->data->dtype);
 		data->byteorder = set->data->byteorder;
 
 		if (set->data->dtype->subtypes) {
@@ -1538,7 +1470,7 @@ out:
 static int list_setelem_cb(struct nftnl_set_elem *nlse, void *arg)
 {
 	struct netlink_ctx *ctx = arg;
-	return netlink_delinearize_setelem(ctx, nlse, ctx->set);
+	return netlink_delinearize_setelem(nlse, ctx->set, &ctx->nft->cache);
 }
 
 static int list_setelem_debug_cb(struct nftnl_set_elem *nlse, void *arg)
@@ -1752,10 +1684,9 @@ struct obj *netlink_delinearize_obj(struct netlink_ctx *ctx,
 		init_list_head(&obj->ct_timeout.timeout_list);
 		obj->ct_timeout.l3proto = nftnl_obj_get_u16(nlo, NFTNL_OBJ_CT_TIMEOUT_L3PROTO);
 		obj->ct_timeout.l4proto = nftnl_obj_get_u8(nlo, NFTNL_OBJ_CT_TIMEOUT_L4PROTO);
-		if (nftnl_obj_is_set(nlo, NFTNL_OBJ_CT_TIMEOUT_ARRAY))
-			memcpy(obj->ct_timeout.timeout,
-			       nftnl_obj_get(nlo, NFTNL_OBJ_CT_TIMEOUT_ARRAY),
-			       NFTNL_CTTIMEOUT_ARRAY_MAX * sizeof(uint32_t));
+		memcpy(obj->ct_timeout.timeout,
+		       nftnl_obj_get(nlo, NFTNL_OBJ_CT_TIMEOUT_ARRAY),
+		       NFTNL_CTTIMEOUT_ARRAY_MAX * sizeof(uint32_t));
 		break;
 	case NFT_OBJECT_LIMIT:
 		obj->limit.rate =
@@ -1789,10 +1720,6 @@ struct obj *netlink_delinearize_obj(struct netlink_ctx *ctx,
 		obj->synproxy.flags =
 			nftnl_obj_get_u32(nlo, NFTNL_OBJ_SYNPROXY_FLAGS);
 		break;
-	default:
-		netlink_io_error(ctx, NULL, "Unknown object type %u", type);
-		obj_free(obj);
-		return NULL;
 	}
 	obj->type = type;
 
@@ -1840,6 +1767,55 @@ int netlink_reset_objs(struct netlink_ctx *ctx, const struct cmd *cmd,
 	return err;
 }
 
+int netlink_reset_rules(struct netlink_ctx *ctx, const struct cmd *cmd,
+		        bool dump)
+{
+	const struct handle *h = &cmd->handle;
+	struct nft_cache_filter f = {
+		.list.table		= h->table.name,
+		.list.chain		= h->chain.name,
+		.list.rule_handle	= h->handle.id,
+	};
+	struct rule *rule, *next, *crule, *cnext;
+	struct table *table;
+	struct chain *chain;
+	int ret;
+
+	ret = rule_cache_dump(ctx, h, &f, dump, true);
+
+	list_for_each_entry_safe(rule, next, &ctx->list, list) {
+		table = table_cache_find(&ctx->nft->cache.table_cache,
+					 rule->handle.table.name,
+					 rule->handle.family);
+		if (!table)
+			continue;
+
+		chain = chain_cache_find(table, rule->handle.chain.name);
+		if (!chain)
+			continue;
+
+		list_del(&rule->list);
+		list_for_each_entry_safe(crule, cnext, &chain->rules, list) {
+			if (crule->handle.handle.id != rule->handle.handle.id)
+				continue;
+
+			list_replace(&crule->list, &rule->list);
+			rule_free(crule);
+			rule = NULL;
+			break;
+		}
+		if (rule) {
+			list_add_tail(&rule->list, &chain->rules);
+		}
+	}
+	list_for_each_entry_safe(rule, next, &ctx->list, list) {
+		list_del(&rule->list);
+		rule_free(rule);
+	}
+
+	return ret;
+}
+
 struct flowtable *
 netlink_delinearize_flowtable(struct netlink_ctx *ctx,
 			      struct nftnl_flowtable *nlo)
@@ -1860,7 +1836,7 @@ netlink_delinearize_flowtable(struct netlink_ctx *ctx,
 	if (nftnl_flowtable_is_set(nlo, NFTNL_FLOWTABLE_FLAGS))
 		flowtable->flags = nftnl_flowtable_get_u32(nlo, NFTNL_FLOWTABLE_FLAGS);
 	dev_array = nftnl_flowtable_get(nlo, NFTNL_FLOWTABLE_DEVICES);
-	while (dev_array && dev_array[len])
+	while (dev_array[len])
 		len++;
 
 	if (len)
@@ -1875,16 +1851,14 @@ netlink_delinearize_flowtable(struct netlink_ctx *ctx,
 		      sizeof(char *), qsort_device_cmp);
 	}
 
-	if (nftnl_flowtable_is_set(nlo, NFTNL_FLOWTABLE_PRIO)) {
-		priority = nftnl_flowtable_get_u32(nlo, NFTNL_FLOWTABLE_PRIO);
-		flowtable->priority.expr =
+	priority = nftnl_flowtable_get_u32(nlo, NFTNL_FLOWTABLE_PRIO);
+	flowtable->priority.expr =
 				constant_expr_alloc(&netlink_location,
 						    &integer_type,
 						    BYTEORDER_HOST_ENDIAN,
 						    sizeof(int) *
 						    BITS_PER_BYTE,
 						    &priority);
-	}
 	flowtable->hook.num =
 		nftnl_flowtable_get_u32(nlo, NFTNL_FLOWTABLE_HOOKNUM);
 	flowtable->flags =
@@ -2091,7 +2065,6 @@ restart:
 		/* Skip unknown and filtered expressions */
 		desc = lhs->payload.desc;
 		if (lhs->dtype == &invalid_type ||
-		    lhs->payload.tmpl == &proto_unknown_template ||
 		    desc->checksum_key == payload_hdr_field(lhs) ||
 		    desc->format.filter & (1 << payload_hdr_field(lhs))) {
 			expr_free(lhs);
@@ -2162,7 +2135,7 @@ static void trace_print_packet(const struct nftnl_trace *nlt,
 				 meta_expr_alloc(&netlink_location,
 						 NFT_META_OIF), octx);
 
-	proto_ctx_init(&ctx, nftnl_trace_get_u32(nlt, NFTNL_TRACE_FAMILY), 0);
+	proto_ctx_init(&ctx, nftnl_trace_get_u32(nlt, NFTNL_TRACE_FAMILY), 0, false);
 	ll_desc = ctx.protocol[PROTO_BASE_LL_HDR].desc;
 	if ((ll_desc == &proto_inet || ll_desc  == &proto_netdev) &&
 	    nftnl_trace_is_set(nlt, NFTNL_TRACE_NFPROTO)) {
