@@ -38,8 +38,6 @@ static bool __expr_cmp(const struct expr *expr_a, const struct expr *expr_b)
 {
 	if (expr_a->etype != expr_b->etype)
 		return false;
-	if (expr_a->len != expr_b->len)
-		return false;
 
 	switch (expr_a->etype) {
 	case EXPR_PAYLOAD:
@@ -49,6 +47,8 @@ static bool __expr_cmp(const struct expr *expr_a, const struct expr *expr_b)
 			return false;
 		if (expr_a->payload.desc != expr_b->payload.desc)
 			return false;
+		if (expr_a->payload.inner_desc != expr_b->payload.inner_desc)
+			return false;
 		if (expr_a->payload.tmpl != expr_b->payload.tmpl)
 			return false;
 		break;
@@ -123,19 +123,7 @@ static bool __expr_cmp(const struct expr *expr_a, const struct expr *expr_b)
 			return false;
 		break;
 	case EXPR_BINOP:
-		if (!__expr_cmp(expr_a->left, expr_b->left))
-			return false;
-
-		return __expr_cmp(expr_a->right, expr_b->right);
-	case EXPR_SYMBOL:
-		if (expr_a->symtype != expr_b->symtype)
-			return false;
-		if (expr_a->symtype != SYMBOL_VALUE)
-			return false;
-
-		return !strcmp(expr_a->identifier, expr_b->identifier);
-	case EXPR_VALUE:
-		return !mpz_cmp(expr_a->value, expr_b->value);
+		return __expr_cmp(expr_a->left, expr_b->left);
 	default:
 		return false;
 	}
@@ -143,25 +131,6 @@ static bool __expr_cmp(const struct expr *expr_a, const struct expr *expr_b)
 	return true;
 }
 
-static bool is_bitmask(const struct expr *expr)
-{
-	switch (expr->etype) {
-	case EXPR_BINOP:
-		if (expr->op == OP_OR &&
-		    !is_bitmask(expr->left))
-			return false;
-
-		return is_bitmask(expr->right);
-	case EXPR_VALUE:
-	case EXPR_SYMBOL:
-		return true;
-	default:
-		break;
-	}
-
-	return false;
-}
-
 static bool stmt_expr_supported(const struct expr *expr)
 {
 	switch (expr->right->etype) {
@@ -172,10 +141,6 @@ static bool stmt_expr_supported(const struct expr *expr)
 	case EXPR_LIST:
 	case EXPR_VALUE:
 		return true;
-	case EXPR_BINOP:
-		if (is_bitmask(expr->right))
-			return true;
-		break;
 	default:
 		break;
 	}
@@ -267,65 +232,28 @@ static bool __stmt_type_eq(const struct stmt *stmt_a, const struct stmt *stmt_b,
 		if (!stmt_a->reject.expr)
 			return true;
 
-		if (!__expr_cmp(stmt_a->reject.expr, stmt_b->reject.expr))
+		if (__expr_cmp(stmt_a->reject.expr, stmt_b->reject.expr))
 			return false;
 		break;
 	case STMT_NAT:
 		if (stmt_a->nat.type != stmt_b->nat.type ||
 		    stmt_a->nat.flags != stmt_b->nat.flags ||
 		    stmt_a->nat.family != stmt_b->nat.family ||
-		    stmt_a->nat.type_flags != stmt_b->nat.type_flags)
+		    stmt_a->nat.type_flags != stmt_b->nat.type_flags ||
+		    (stmt_a->nat.addr &&
+		     stmt_a->nat.addr->etype != EXPR_SYMBOL &&
+		     stmt_a->nat.addr->etype != EXPR_RANGE) ||
+		    (stmt_b->nat.addr &&
+		     stmt_b->nat.addr->etype != EXPR_SYMBOL &&
+		     stmt_b->nat.addr->etype != EXPR_RANGE) ||
+		    (stmt_a->nat.proto &&
+		     stmt_a->nat.proto->etype != EXPR_SYMBOL &&
+		     stmt_a->nat.proto->etype != EXPR_RANGE) ||
+		    (stmt_b->nat.proto &&
+		     stmt_b->nat.proto->etype != EXPR_SYMBOL &&
+		     stmt_b->nat.proto->etype != EXPR_RANGE))
 			return false;
 
-		switch (stmt_a->nat.type) {
-		case NFT_NAT_SNAT:
-		case NFT_NAT_DNAT:
-			if ((stmt_a->nat.addr &&
-			     stmt_a->nat.addr->etype != EXPR_SYMBOL &&
-			     stmt_a->nat.addr->etype != EXPR_RANGE) ||
-			    (stmt_b->nat.addr &&
-			     stmt_b->nat.addr->etype != EXPR_SYMBOL &&
-			     stmt_b->nat.addr->etype != EXPR_RANGE) ||
-			    (stmt_a->nat.proto &&
-			     stmt_a->nat.proto->etype != EXPR_SYMBOL &&
-			     stmt_a->nat.proto->etype != EXPR_RANGE) ||
-			    (stmt_b->nat.proto &&
-			     stmt_b->nat.proto->etype != EXPR_SYMBOL &&
-			     stmt_b->nat.proto->etype != EXPR_RANGE))
-				return false;
-			break;
-		case NFT_NAT_MASQ:
-			break;
-		case NFT_NAT_REDIR:
-			if ((stmt_a->nat.proto &&
-			     stmt_a->nat.proto->etype != EXPR_SYMBOL &&
-			     stmt_a->nat.proto->etype != EXPR_RANGE) ||
-			    (stmt_b->nat.proto &&
-			     stmt_b->nat.proto->etype != EXPR_SYMBOL &&
-			     stmt_b->nat.proto->etype != EXPR_RANGE))
-				return false;
-
-			/* it should be possible to infer implicit redirections
-			 * such as:
-			 *
-			 *	tcp dport 1234 redirect
-			 *	tcp dport 3456 redirect to :7890
-			 * merge:
-			 *	redirect to tcp dport map { 1234 : 1234, 3456 : 7890 }
-			 *
-			 * currently not implemented.
-			 */
-			if (fully_compare &&
-			    stmt_a->nat.type == NFT_NAT_REDIR &&
-			    stmt_b->nat.type == NFT_NAT_REDIR &&
-			    (!!stmt_a->nat.proto ^ !!stmt_b->nat.proto))
-				return false;
-
-			break;
-		default:
-			assert(0);
-		}
-
 		return true;
 	default:
 		/* ... Merging anything else is yet unsupported. */
@@ -444,11 +372,9 @@ static int rule_collect_stmts(struct optimize_ctx *ctx, struct rule *rule)
 			break;
 		case STMT_NAT:
 			if ((stmt->nat.addr &&
-			     (stmt->nat.addr->etype == EXPR_MAP ||
-			      stmt->nat.addr->etype == EXPR_VARIABLE)) ||
+			     stmt->nat.addr->etype == EXPR_MAP) ||
 			    (stmt->nat.proto &&
-			     (stmt->nat.proto->etype == EXPR_MAP ||
-			      stmt->nat.proto->etype == EXPR_VARIABLE))) {
+			     stmt->nat.proto->etype == EXPR_MAP)) {
 				clone->ops = &unsupported_stmt_ops;
 				break;
 			}
@@ -549,8 +475,6 @@ struct merge {
 	/* statements to be merged (index relative to statement matrix) */
 	uint32_t	stmt[MAX_STMTS];
 	uint32_t	num_stmts;
-	/* merge has been invalidated */
-	bool		skip;
 };
 
 static void merge_expr_stmts(const struct optimize_ctx *ctx,
@@ -668,16 +592,6 @@ static void __merge_concat(const struct optimize_ctx *ctx, uint32_t i,
 				clone = expr_clone(stmt_a->expr->right);
 				compound_expr_add(concat, clone);
 				break;
-			case EXPR_LIST:
-				list_for_each_entry(expr, &stmt_a->expr->right->expressions, list) {
-					concat_clone = expr_clone(concat);
-					clone = expr_clone(expr);
-					compound_expr_add(concat_clone, clone);
-					list_add_tail(&concat_clone->list, &pending_list);
-				}
-				list_del(&concat->list);
-				expr_free(concat);
-				break;
 			default:
 				assert(0);
 				break;
@@ -738,40 +652,26 @@ static void merge_concat_stmts(const struct optimize_ctx *ctx,
 	}
 }
 
-static void build_verdict_map(struct expr *expr, struct stmt *verdict,
-			      struct expr *set, struct stmt *counter)
+static void build_verdict_map(struct expr *expr, struct stmt *verdict, struct expr *set)
 {
 	struct expr *item, *elem, *mapping;
-	struct stmt *counter_elem;
 
 	switch (expr->etype) {
 	case EXPR_LIST:
 		list_for_each_entry(item, &expr->expressions, list) {
 			elem = set_elem_expr_alloc(&internal_location, expr_get(item));
-			if (counter) {
-				counter_elem = counter_stmt_alloc(&counter->location);
-				list_add_tail(&counter_elem->list, &elem->stmt_list);
-			}
-
 			mapping = mapping_expr_alloc(&internal_location, elem,
 						     expr_get(verdict->expr));
 			compound_expr_add(set, mapping);
 		}
-		stmt_free(counter);
 		break;
 	case EXPR_SET:
 		list_for_each_entry(item, &expr->expressions, list) {
 			elem = set_elem_expr_alloc(&internal_location, expr_get(item->key));
-			if (counter) {
-				counter_elem = counter_stmt_alloc(&counter->location);
-				list_add_tail(&counter_elem->list, &elem->stmt_list);
-			}
-
 			mapping = mapping_expr_alloc(&internal_location, elem,
 						     expr_get(verdict->expr));
 			compound_expr_add(set, mapping);
 		}
-		stmt_free(counter);
 		break;
 	case EXPR_PREFIX:
 	case EXPR_RANGE:
@@ -779,9 +679,6 @@ static void build_verdict_map(struct expr *expr, struct stmt *verdict,
 	case EXPR_SYMBOL:
 	case EXPR_CONCAT:
 		elem = set_elem_expr_alloc(&internal_location, expr_get(expr));
-		if (counter)
-			list_add_tail(&counter->list, &elem->stmt_list);
-
 		mapping = mapping_expr_alloc(&internal_location, elem,
 					     expr_get(verdict->expr));
 		compound_expr_add(set, mapping);
@@ -810,26 +707,6 @@ static void remove_counter(const struct optimize_ctx *ctx, uint32_t from)
 	}
 }
 
-static struct stmt *zap_counter(const struct optimize_ctx *ctx, uint32_t from)
-{
-	struct stmt *stmt;
-	uint32_t i;
-
-	/* remove counter statement */
-	for (i = 0; i < ctx->num_stmts; i++) {
-		stmt = ctx->stmt_matrix[from][i];
-		if (!stmt)
-			continue;
-
-		if (stmt->ops->type == STMT_COUNTER) {
-			list_del(&stmt->list);
-			return stmt;
-		}
-	}
-
-	return NULL;
-}
-
 static void merge_stmts_vmap(const struct optimize_ctx *ctx,
 			     uint32_t from, uint32_t to,
 			     const struct merge *merge)
@@ -837,33 +714,31 @@ static void merge_stmts_vmap(const struct optimize_ctx *ctx,
 	struct stmt *stmt_a = ctx->stmt_matrix[from][merge->stmt[0]];
 	struct stmt *stmt_b, *verdict_a, *verdict_b, *stmt;
 	struct expr *expr_a, *expr_b, *expr, *left, *set;
-	struct stmt *counter;
 	uint32_t i;
 	int k;
 
 	k = stmt_verdict_find(ctx);
 	assert(k >= 0);
 
+	verdict_a = ctx->stmt_matrix[from][k];
 	set = set_expr_alloc(&internal_location, NULL);
 	set->set_flags |= NFT_SET_ANONYMOUS;
 
 	expr_a = stmt_a->expr->right;
-	verdict_a = ctx->stmt_matrix[from][k];
-	counter = zap_counter(ctx, from);
-	build_verdict_map(expr_a, verdict_a, set, counter);
-
+	build_verdict_map(expr_a, verdict_a, set);
 	for (i = from + 1; i <= to; i++) {
 		stmt_b = ctx->stmt_matrix[i][merge->stmt[0]];
 		expr_b = stmt_b->expr->right;
 		verdict_b = ctx->stmt_matrix[i][k];
-		counter = zap_counter(ctx, i);
-		build_verdict_map(expr_b, verdict_b, set, counter);
+
+		build_verdict_map(expr_b, verdict_b, set);
 	}
 
 	left = expr_get(stmt_a->expr->left);
 	expr = map_expr_alloc(&internal_location, left, set);
 	stmt = verdict_stmt_alloc(&internal_location, expr);
 
+	remove_counter(ctx, from);
 	list_add(&stmt->list, &stmt_a->list);
 	list_del(&stmt_a->list);
 	stmt_free(stmt_a);
@@ -876,25 +751,17 @@ static void __merge_concat_stmts_vmap(const struct optimize_ctx *ctx,
 				      struct expr *set, struct stmt *verdict)
 {
 	struct expr *concat, *next, *elem, *mapping;
-	struct stmt *counter, *counter_elem;
 	LIST_HEAD(concat_list);
 
-	counter = zap_counter(ctx, i);
 	__merge_concat(ctx, i, merge, &concat_list);
 
 	list_for_each_entry_safe(concat, next, &concat_list, list) {
 		list_del(&concat->list);
 		elem = set_elem_expr_alloc(&internal_location, concat);
-		if (counter) {
-			counter_elem = counter_stmt_alloc(&counter->location);
-			list_add_tail(&counter_elem->list, &elem->stmt_list);
-		}
-
 		mapping = mapping_expr_alloc(&internal_location, elem,
 					     expr_get(verdict->expr));
 		compound_expr_add(set, mapping);
 	}
-	stmt_free(counter);
 }
 
 static void merge_concat_stmts_vmap(const struct optimize_ctx *ctx,
@@ -929,6 +796,7 @@ static void merge_concat_stmts_vmap(const struct optimize_ctx *ctx,
 	expr = map_expr_alloc(&internal_location, concat_a, set);
 	stmt = verdict_stmt_alloc(&internal_location, expr);
 
+	remove_counter(ctx, from);
 	list_add(&stmt->list, &orig_stmt->list);
 	list_del(&orig_stmt->list);
 	stmt_free(orig_stmt);
@@ -969,35 +837,12 @@ static bool stmt_verdict_cmp(const struct optimize_ctx *ctx,
 	return true;
 }
 
-static int stmt_nat_type(const struct optimize_ctx *ctx, int from,
-			 enum nft_nat_etypes *nat_type)
+static int stmt_nat_find(const struct optimize_ctx *ctx)
 {
-	uint32_t j;
-
-	for (j = 0; j < ctx->num_stmts; j++) {
-		if (!ctx->stmt_matrix[from][j])
-			continue;
-
-		if (ctx->stmt_matrix[from][j]->ops->type == STMT_NAT) {
-			*nat_type = ctx->stmt_matrix[from][j]->nat.type;
-			return 0;
-		}
-	}
-
-	return -1;
-}
-
-static int stmt_nat_find(const struct optimize_ctx *ctx, int from)
-{
-	enum nft_nat_etypes nat_type;
 	uint32_t i;
 
-	if (stmt_nat_type(ctx, from, &nat_type) < 0)
-		return -1;
-
 	for (i = 0; i < ctx->num_stmts; i++) {
-		if (ctx->stmt[i]->ops->type != STMT_NAT ||
-		    ctx->stmt[i]->nat.type != nat_type)
+		if (ctx->stmt[i]->ops->type != STMT_NAT)
 			continue;
 
 		return i;
@@ -1010,24 +855,16 @@ static struct expr *stmt_nat_expr(struct stmt *nat_stmt)
 {
 	struct expr *nat_expr;
 
-	assert(nat_stmt->ops->type == STMT_NAT);
-
 	if (nat_stmt->nat.proto) {
-		if (nat_stmt->nat.addr) {
-			nat_expr = concat_expr_alloc(&internal_location);
-			compound_expr_add(nat_expr, expr_get(nat_stmt->nat.addr));
-			compound_expr_add(nat_expr, expr_get(nat_stmt->nat.proto));
-		} else {
-			nat_expr = expr_get(nat_stmt->nat.proto);
-		}
+		nat_expr = concat_expr_alloc(&internal_location);
+		compound_expr_add(nat_expr, expr_get(nat_stmt->nat.addr));
+		compound_expr_add(nat_expr, expr_get(nat_stmt->nat.proto));
 		expr_free(nat_stmt->nat.proto);
 		nat_stmt->nat.proto = NULL;
 	} else {
 		nat_expr = expr_get(nat_stmt->nat.addr);
 	}
 
-	assert(nat_expr);
-
 	return nat_expr;
 }
 
@@ -1040,7 +877,7 @@ static void merge_nat(const struct optimize_ctx *ctx,
 	struct stmt *stmt, *nat_stmt;
 	uint32_t i;
 
-	k = stmt_nat_find(ctx, from);
+	k = stmt_nat_find(ctx);
 	assert(k >= 0);
 
 	set = set_expr_alloc(&internal_location, NULL);
@@ -1073,10 +910,7 @@ static void merge_nat(const struct optimize_ctx *ctx,
 		nat_stmt->nat.family = family;
 
 	expr_free(nat_stmt->nat.addr);
-	if (nat_stmt->nat.type == NFT_NAT_REDIR)
-		nat_stmt->nat.proto = expr;
-	else
-		nat_stmt->nat.addr = expr;
+	nat_stmt->nat.addr = expr;
 
 	remove_counter(ctx, from);
 	list_del(&stmt->list);
@@ -1092,7 +926,7 @@ static void merge_concat_nat(const struct optimize_ctx *ctx,
 	struct stmt *stmt, *nat_stmt;
 	uint32_t i, j;
 
-	k = stmt_nat_find(ctx, from);
+	k = stmt_nat_find(ctx);
 	assert(k >= 0);
 
 	set = set_expr_alloc(&internal_location, NULL);
@@ -1177,36 +1011,22 @@ static void rule_optimize_print(struct output_ctx *octx,
 	fprintf(octx->error_fp, "%s\n", line);
 }
 
-enum {
-	MERGE_BY_VERDICT,
-	MERGE_BY_NAT_MAP,
-	MERGE_BY_NAT,
-};
-
-static uint32_t merge_stmt_type(const struct optimize_ctx *ctx,
-				uint32_t from, uint32_t to)
+static enum stmt_types merge_stmt_type(const struct optimize_ctx *ctx,
+				       uint32_t from, uint32_t to)
 {
-	const struct stmt *stmt;
 	uint32_t i, j;
 
 	for (i = from; i <= to; i++) {
 		for (j = 0; j < ctx->num_stmts; j++) {
-			stmt = ctx->stmt_matrix[i][j];
-			if (!stmt)
+			if (!ctx->stmt_matrix[i][j])
 				continue;
-			if (stmt->ops->type == STMT_NAT) {
-				if ((stmt->nat.type == NFT_NAT_REDIR &&
-				     !stmt->nat.proto) ||
-				    stmt->nat.type == NFT_NAT_MASQ)
-					return MERGE_BY_NAT;
-
-				return MERGE_BY_NAT_MAP;
-			}
+			if (ctx->stmt_matrix[i][j]->ops->type == STMT_NAT)
+				return STMT_NAT;
 		}
 	}
 
 	/* merge by verdict, even if no verdict is specified. */
-	return MERGE_BY_VERDICT;
+	return STMT_VERDICT;
 }
 
 static void merge_rules(const struct optimize_ctx *ctx,
@@ -1214,14 +1034,14 @@ static void merge_rules(const struct optimize_ctx *ctx,
 			const struct merge *merge,
 			struct output_ctx *octx)
 {
-	uint32_t merge_type;
+	enum stmt_types stmt_type;
 	bool same_verdict;
 	uint32_t i;
 
-	merge_type = merge_stmt_type(ctx, from, to);
+	stmt_type = merge_stmt_type(ctx, from, to);
 
-	switch (merge_type) {
-	case MERGE_BY_VERDICT:
+	switch (stmt_type) {
+	case STMT_VERDICT:
 		same_verdict = stmt_verdict_cmp(ctx, from, to);
 		if (merge->num_stmts > 1) {
 			if (same_verdict)
@@ -1235,18 +1055,12 @@ static void merge_rules(const struct optimize_ctx *ctx,
 				merge_stmts_vmap(ctx, from, to, merge);
 		}
 		break;
-	case MERGE_BY_NAT_MAP:
+	case STMT_NAT:
 		if (merge->num_stmts > 1)
 			merge_concat_nat(ctx, from, to, merge);
 		else
 			merge_nat(ctx, from, to, merge);
 		break;
-	case MERGE_BY_NAT:
-		if (merge->num_stmts > 1)
-			merge_concat_stmts(ctx, from, to, merge);
-		else
-			merge_stmts(ctx, from, to, merge);
-		break;
 	default:
 		assert(0);
 	}
@@ -1380,42 +1194,8 @@ static int chain_optimize(struct nft_ctx *nft, struct list_head *rules)
 		}
 	}
 
-	/* Step 4: Invalidate merge in case of duplicated keys in set/map. */
+	/* Step 4: Infer how to merge the candidate rules */
 	for (k = 0; k < num_merges; k++) {
-		uint32_t r1, r2;
-
-		i = merge[k].rule_from;
-
-		for (r1 = i; r1 < i + merge[k].num_rules; r1++) {
-			for (r2 = r1 + 1; r2 < i + merge[k].num_rules; r2++) {
-				bool match_same_value = true, match_seen = false;
-
-				for (m = 0; m < ctx->num_stmts; m++) {
-					if (!ctx->stmt_matrix[r1][m])
-						continue;
-
-					switch (ctx->stmt_matrix[r1][m]->ops->type) {
-					case STMT_EXPRESSION:
-						match_seen = true;
-						if (!__expr_cmp(ctx->stmt_matrix[r1][m]->expr->right,
-							        ctx->stmt_matrix[r2][m]->expr->right))
-							match_same_value = false;
-						break;
-					default:
-						break;
-					}
-				}
-				if (match_seen && match_same_value)
-					merge[k].skip = true;
-			}
-		}
-	}
-
-	/* Step 5: Infer how to merge the candidate rules */
-	for (k = 0; k < num_merges; k++) {
-		if (merge[k].skip)
-			continue;
-
 		i = merge[k].rule_from;
 
 		for (m = 0; m < ctx->num_stmts; m++) {
