@@ -210,9 +210,6 @@ struct set *set_lookup_fuzzy(const char *set_name,
 	struct table *table;
 	struct set *set;
 
-	if (!set_name)
-		return NULL;
-
 	string_misspell_init(&st);
 
 	list_for_each_entry(table, &cache->table_cache.list, cache.list) {
@@ -730,6 +727,7 @@ void chain_free(struct chain *chain)
 	list_for_each_entry_safe(rule, next, &chain->rules, list)
 		rule_free(rule);
 	handle_free(&chain->handle);
+	scope_release(&chain->scope);
 	xfree(chain->type.str);
 	expr_free(chain->dev_expr);
 	for (i = 0; i < chain->dev_array_len; i++)
@@ -738,11 +736,6 @@ void chain_free(struct chain *chain)
 	expr_free(chain->priority.expr);
 	expr_free(chain->policy);
 	xfree(chain->comment);
-
-	/* MUST be released after all expressions, they could
-	 * hold refcounts.
-	 */
-	scope_release(&chain->scope);
 	xfree(chain);
 }
 
@@ -871,7 +864,7 @@ struct prio_tag {
 	const char *str;
 };
 
-static const struct prio_tag std_prios[] = {
+const static struct prio_tag std_prios[] = {
 	{ NF_IP_PRI_RAW,      "raw" },
 	{ NF_IP_PRI_MANGLE,   "mangle" },
 	{ NF_IP_PRI_NAT_DST,  "dstnat" },
@@ -880,7 +873,7 @@ static const struct prio_tag std_prios[] = {
 	{ NF_IP_PRI_NAT_SRC,  "srcnat" },
 };
 
-static const struct prio_tag bridge_std_prios[] = {
+const static struct prio_tag bridge_std_prios[] = {
 	{ NF_BR_PRI_NAT_DST_BRIDGED,  "dstnat" },
 	{ NF_BR_PRI_FILTER_BRIDGED,   "filter" },
 	{ NF_BR_PRI_NAT_DST_OTHER,    "out" },
@@ -934,8 +927,7 @@ static bool std_prio_family_hook_compat(int prio, int family, int hook)
 		case NFPROTO_INET:
 		case NFPROTO_IPV4:
 		case NFPROTO_IPV6:
-			if (hook == NF_INET_PRE_ROUTING ||
-			    hook == NF_INET_LOCAL_OUT)
+			if (hook == NF_INET_PRE_ROUTING)
 				return true;
 		}
 		break;
@@ -944,8 +936,7 @@ static bool std_prio_family_hook_compat(int prio, int family, int hook)
 		case NFPROTO_INET:
 		case NFPROTO_IPV4:
 		case NFPROTO_IPV6:
-			if (hook == NF_INET_LOCAL_IN ||
-			    hook == NF_INET_POST_ROUTING)
+			if (hook == NF_INET_POST_ROUTING)
 				return true;
 		}
 	}
@@ -978,11 +969,10 @@ static const char *prio2str(const struct output_ctx *octx,
 			    const struct expr *expr)
 {
 	const struct prio_tag *prio_arr;
-	const uint32_t reach = 10;
+	int std_prio, offset, prio;
 	const char *std_prio_str;
-	int std_prio, prio;
+	const int reach = 10;
 	size_t i, arr_size;
-	int64_t offset;
 
 	mpz_export_data(&prio, expr->value, BYTEORDER_HOST_ENDIAN, sizeof(int));
 	if (family == NFPROTO_BRIDGE) {
@@ -997,21 +987,19 @@ static const char *prio2str(const struct output_ctx *octx,
 		for (i = 0; i < arr_size; ++i) {
 			std_prio = prio_arr[i].val;
 			std_prio_str = prio_arr[i].str;
-
-			offset = (int64_t)prio - std_prio;
-			if (llabs(offset) <= reach) {
+			if (abs(prio - std_prio) <= reach) {
 				if (!std_prio_family_hook_compat(std_prio,
 								 family, hook))
 					break;
-
+				offset = prio - std_prio;
 				strncpy(buf, std_prio_str, bufsize);
 				if (offset > 0)
 					snprintf(buf + strlen(buf),
-						 bufsize - strlen(buf), " + %" PRIu64,
+						 bufsize - strlen(buf), " + %d",
 						 offset);
 				else if (offset < 0)
 					snprintf(buf + strlen(buf),
-						 bufsize - strlen(buf), " - %" PRIu64,
+						 bufsize - strlen(buf), " - %d",
 						 -offset);
 				return buf;
 			}
@@ -1037,27 +1025,23 @@ static void chain_print_declaration(const struct chain *chain,
 		nft_print(octx, "\n\t\tcomment \"%s\"", chain->comment);
 	nft_print(octx, "\n");
 	if (chain->flags & CHAIN_F_BASECHAIN) {
-		if (chain->type.str)
-			nft_print(octx, "\t\ttype %s hook %s", chain->type.str,
-				  hooknum2str(chain->handle.family, chain->hook.num));
-
+		nft_print(octx, "\t\ttype %s hook %s", chain->type.str,
+			  hooknum2str(chain->handle.family, chain->hook.num));
 		if (chain->dev_array_len == 1) {
 			nft_print(octx, " device \"%s\"", chain->dev_array[0]);
 		} else if (chain->dev_array_len > 1) {
 			nft_print(octx, " devices = { ");
 			for (i = 0; i < chain->dev_array_len; i++) {
-				nft_print(octx, "\"%s\"", chain->dev_array[i]);
+				nft_print(octx, "%s", chain->dev_array[i]);
 					if (i + 1 != chain->dev_array_len)
 						nft_print(octx, ", ");
 			}
 			nft_print(octx, " }");
 		}
-
-		if (chain->priority.expr)
-			nft_print(octx, " priority %s;",
-				  prio2str(octx, priobuf, sizeof(priobuf),
-					   chain->handle.family, chain->hook.num,
-					   chain->priority.expr));
+		nft_print(octx, " priority %s;",
+			  prio2str(octx, priobuf, sizeof(priobuf),
+				   chain->handle.family, chain->hook.num,
+				   chain->priority.expr));
 		if (chain->policy) {
 			mpz_export_data(&policy, chain->policy->value,
 					BYTEORDER_HOST_ENDIAN, sizeof(int));
@@ -1194,9 +1178,6 @@ struct table *table_lookup_fuzzy(const struct handle *h,
 	struct string_misspell_state st;
 	struct table *table;
 
-	if (!h->table.name)
-		return NULL;
-
 	string_misspell_init(&st);
 
 	list_for_each_entry(table, &cache->table_cache.list, cache.list) {
@@ -1317,200 +1298,6 @@ struct cmd *cmd_alloc(enum cmd_ops op, enum cmd_obj obj,
 	return cmd;
 }
 
-void cmd_add_loc(struct cmd *cmd, uint16_t offset, const struct location *loc)
-{
-	if (cmd->num_attrs >= cmd->attr_array_len) {
-		cmd->attr_array_len *= 2;
-		cmd->attr = xrealloc(cmd->attr, sizeof(struct nlerr_loc) * cmd->attr_array_len);
-	}
-
-	cmd->attr[cmd->num_attrs].offset = offset;
-	cmd->attr[cmd->num_attrs].location = loc;
-	cmd->num_attrs++;
-}
-
-static void nft_cmd_expand_chain(struct chain *chain, struct list_head *new_cmds)
-{
-	struct rule *rule, *next;
-	struct handle h;
-	struct cmd *new;
-
-	list_for_each_entry_safe(rule, next, &chain->rules, list) {
-		list_del(&rule->list);
-		handle_merge(&rule->handle, &chain->handle);
-		memset(&h, 0, sizeof(h));
-		handle_merge(&h, &chain->handle);
-		if (chain->flags & CHAIN_F_BINDING) {
-			rule->handle.chain_id = chain->handle.chain_id;
-			rule->handle.chain.location = chain->location;
-		}
-		new = cmd_alloc(CMD_ADD, CMD_OBJ_RULE, &h,
-				&rule->location, rule);
-		list_add_tail(&new->list, new_cmds);
-	}
-}
-
-void nft_cmd_expand(struct cmd *cmd)
-{
-	struct list_head new_cmds;
-	struct flowtable *ft;
-	struct table *table;
-	struct chain *chain;
-	struct set *set;
-	struct obj *obj;
-	struct cmd *new;
-	struct handle h;
-
-	init_list_head(&new_cmds);
-
-	switch (cmd->obj) {
-	case CMD_OBJ_TABLE:
-		table = cmd->table;
-		if (!table)
-			return;
-
-		list_for_each_entry(chain, &table->chains, list) {
-			handle_merge(&chain->handle, &table->handle);
-			memset(&h, 0, sizeof(h));
-			handle_merge(&h, &chain->handle);
-			h.chain_id = chain->handle.chain_id;
-			new = cmd_alloc(CMD_ADD, CMD_OBJ_CHAIN, &h,
-					&chain->location, chain_get(chain));
-			list_add_tail(&new->list, &new_cmds);
-		}
-		list_for_each_entry(obj, &table->objs, list) {
-			handle_merge(&obj->handle, &table->handle);
-			memset(&h, 0, sizeof(h));
-			handle_merge(&h, &obj->handle);
-			new = cmd_alloc(CMD_ADD, obj_type_to_cmd(obj->type), &h,
-					&obj->location, obj_get(obj));
-			list_add_tail(&new->list, &new_cmds);
-		}
-		list_for_each_entry(set, &table->sets, list) {
-			handle_merge(&set->handle, &table->handle);
-			memset(&h, 0, sizeof(h));
-			handle_merge(&h, &set->handle);
-			new = cmd_alloc(CMD_ADD, CMD_OBJ_SET, &h,
-					&set->location, set_get(set));
-			list_add_tail(&new->list, &new_cmds);
-		}
-		list_for_each_entry(ft, &table->flowtables, list) {
-			handle_merge(&ft->handle, &table->handle);
-			memset(&h, 0, sizeof(h));
-			handle_merge(&h, &ft->handle);
-			new = cmd_alloc(CMD_ADD, CMD_OBJ_FLOWTABLE, &h,
-					&ft->location, flowtable_get(ft));
-			list_add_tail(&new->list, &new_cmds);
-		}
-		list_for_each_entry(chain, &table->chains, list)
-			nft_cmd_expand_chain(chain, &new_cmds);
-
-		list_splice(&new_cmds, &cmd->list);
-		break;
-	case CMD_OBJ_CHAIN:
-		chain = cmd->chain;
-		if (!chain || list_empty(&chain->rules))
-			break;
-
-		nft_cmd_expand_chain(chain, &new_cmds);
-		list_splice(&new_cmds, &cmd->list);
-		break;
-	case CMD_OBJ_SET:
-	case CMD_OBJ_MAP:
-		set = cmd->set;
-		if (!set->init)
-			break;
-
-		memset(&h, 0, sizeof(h));
-		handle_merge(&h, &set->handle);
-		new = cmd_alloc(CMD_ADD, CMD_OBJ_SETELEMS, &h,
-				&set->location, set_get(set));
-		list_add(&new->list, &cmd->list);
-		break;
-	default:
-		break;
-	}
-}
-
-bool nft_cmd_collapse(struct list_head *cmds)
-{
-	struct cmd *cmd, *next, *elems = NULL;
-	struct expr *expr, *enext;
-	bool collapse = false;
-
-	list_for_each_entry_safe(cmd, next, cmds, list) {
-		if (cmd->op != CMD_ADD &&
-		    cmd->op != CMD_CREATE) {
-			elems = NULL;
-			continue;
-		}
-
-		if (cmd->obj != CMD_OBJ_ELEMENTS) {
-			elems = NULL;
-			continue;
-		}
-
-		if (cmd->expr->etype == EXPR_VARIABLE)
-			continue;
-
-		if (!elems) {
-			elems = cmd;
-			continue;
-		}
-
-		if (cmd->op != elems->op) {
-			elems = cmd;
-			continue;
-		}
-
-		if (elems->handle.family != cmd->handle.family ||
-		    strcmp(elems->handle.table.name, cmd->handle.table.name) ||
-		    strcmp(elems->handle.set.name, cmd->handle.set.name)) {
-			elems = cmd;
-			continue;
-		}
-
-		collapse = true;
-		list_for_each_entry_safe(expr, enext, &cmd->expr->expressions, list) {
-			expr->cmd = cmd;
-			list_move_tail(&expr->list, &elems->expr->expressions);
-		}
-		elems->expr->size += cmd->expr->size;
-		list_move_tail(&cmd->list, &elems->collapse_list);
-	}
-
-	return collapse;
-}
-
-void nft_cmd_uncollapse(struct list_head *cmds)
-{
-	struct cmd *cmd, *cmd_next, *collapse_cmd, *collapse_cmd_next;
-	struct expr *expr, *next;
-
-	list_for_each_entry_safe(cmd, cmd_next, cmds, list) {
-		if (list_empty(&cmd->collapse_list))
-			continue;
-
-		assert(cmd->obj == CMD_OBJ_ELEMENTS);
-
-		list_for_each_entry_safe(expr, next, &cmd->expr->expressions, list) {
-			if (!expr->cmd)
-				continue;
-
-			list_move_tail(&expr->list, &expr->cmd->expr->expressions);
-			cmd->expr->size--;
-			expr->cmd = NULL;
-		}
-
-		list_for_each_entry_safe(collapse_cmd, collapse_cmd_next, &cmd->collapse_list, list) {
-			if (cmd->elem.set)
-				collapse_cmd->elem.set = set_get(cmd->elem.set);
-
-			list_add(&collapse_cmd->list, &cmd->list);
-		}
-	}
-}
-
 struct markup *markup_alloc(uint32_t format)
 {
 	struct markup *markup;
@@ -1547,9 +1334,6 @@ void monitor_free(struct monitor *m)
 
 void cmd_free(struct cmd *cmd)
 {
-	if (cmd == NULL)
-		return;
-
 	handle_free(&cmd->handle);
 	if (cmd->data != NULL) {
 		switch (cmd->obj) {
@@ -1651,13 +1435,7 @@ static int do_add_set(struct netlink_ctx *ctx, struct cmd *cmd,
 			return -1;
 	}
 
-	if (mnl_nft_set_add(ctx, cmd, flags) < 0)
-		return -1;
-
-	if (set_is_anonymous(set->flags))
-		return __do_add_elements(ctx, cmd, set, set->init, flags);
-
-	return 0;
+	return mnl_nft_set_add(ctx, cmd, flags);
 }
 
 static int do_command_add(struct netlink_ctx *ctx, struct cmd *cmd, bool excl)
@@ -1784,6 +1562,11 @@ static int do_list_table(struct netlink_ctx *ctx, struct cmd *cmd,
 
 static int do_list_sets(struct netlink_ctx *ctx, struct cmd *cmd)
 {
+	struct print_fmt_options opts = {
+		.tab		= "\t",
+		.nl		= "\n",
+		.stmt_separator	= "\n",
+	};
 	struct table *table;
 	struct set *set;
 
@@ -1806,7 +1589,8 @@ static int do_list_sets(struct netlink_ctx *ctx, struct cmd *cmd)
 			if (cmd->obj == CMD_OBJ_MAPS &&
 			    !map_is_literal(set->flags))
 				continue;
-			set_print(set, &ctx->nft->output);
+			set_print_declaration(set, &opts, &ctx->nft->output);
+			nft_print(&ctx->nft->output, "%s}%s", opts.tab, opts.nl);
 		}
 
 		nft_print(&ctx->nft->output, "}\n");
@@ -1858,9 +1642,6 @@ struct obj *obj_lookup_fuzzy(const char *obj_name,
 	struct table *table;
 	struct obj *obj;
 
-	if (!obj_name)
-		return NULL;
-
 	string_misspell_init(&st);
 
 	list_for_each_entry(table, &cache->table_cache.list, cache.list) {
@@ -1893,14 +1674,11 @@ static void print_proto_timeout_policy(uint8_t l4, const uint32_t *timeout,
 	nft_print(octx, "%s%spolicy = { ", opts->tab, opts->tab);
 	for (i = 0; i < timeout_protocol[l4].array_size; i++) {
 		if (timeout[i] != timeout_protocol[l4].dflt_timeout[i]) {
-			uint64_t timeout_ms;
-
 			if (comma)
 				nft_print(octx, ", ");
-			timeout_ms = timeout[i] * 1000u;
-			nft_print(octx, "%s : ",
-				  timeout_protocol[l4].state_to_name[i]);
-			time_print(timeout_ms, octx);
+			nft_print(octx, "%s : %u",
+				  timeout_protocol[l4].state_to_name[i],
+				  timeout[i]);
 			comma = true;
 		}
 	}
@@ -2284,20 +2062,17 @@ static void flowtable_print_declaration(const struct flowtable *flowtable,
 	if (nft_output_handle(octx))
 		nft_print(octx, " # handle %" PRIu64, flowtable->handle.handle.id);
 	nft_print(octx, "%s", opts->nl);
-
-	if (flowtable->priority.expr) {
-		nft_print(octx, "%s%shook %s priority %s%s",
-			  opts->tab, opts->tab,
-			  hooknum2str(NFPROTO_NETDEV, flowtable->hook.num),
-			  prio2str(octx, priobuf, sizeof(priobuf), NFPROTO_NETDEV,
-				   flowtable->hook.num, flowtable->priority.expr),
-			  opts->stmt_separator);
-	}
+	nft_print(octx, "%s%shook %s priority %s%s",
+		  opts->tab, opts->tab,
+		  hooknum2str(NFPROTO_NETDEV, flowtable->hook.num),
+		  prio2str(octx, priobuf, sizeof(priobuf), NFPROTO_NETDEV,
+			   flowtable->hook.num, flowtable->priority.expr),
+		  opts->stmt_separator);
 
 	if (flowtable->dev_array_len > 0) {
 		nft_print(octx, "%s%sdevices = { ", opts->tab, opts->tab);
 		for (i = 0; i < flowtable->dev_array_len; i++) {
-			nft_print(octx, "\"%s\"", flowtable->dev_array[i]);
+			nft_print(octx, "%s", flowtable->dev_array[i]);
 			if (i + 1 != flowtable->dev_array_len)
 				nft_print(octx, ", ");
 		}
@@ -2332,21 +2107,6 @@ void flowtable_print(const struct flowtable *s, struct output_ctx *octx)
 	do_flowtable_print(s, &opts, octx);
 }
 
-void flowtable_print_plain(const struct flowtable *ft, struct output_ctx *octx)
-{
-	struct print_fmt_options opts = {
-		.tab		= "",
-		.nl		= " ",
-		.table		= ft->handle.table.name,
-		.family		= family2str(ft->handle.family),
-		.stmt_separator = "; ",
-	};
-
-	flowtable_print_declaration(ft, &opts, octx);
-	nft_print(octx, "}");
-}
-
-
 struct flowtable *flowtable_lookup_fuzzy(const char *ft_name,
 					 const struct nft_cache *cache,
 					 const struct table **t)
@@ -2355,9 +2115,6 @@ struct flowtable *flowtable_lookup_fuzzy(const char *ft_name,
 	struct table *table;
 	struct flowtable *ft;
 
-	if (!ft_name)
-		return NULL;
-
 	string_misspell_init(&st);
 
 	list_for_each_entry(table, &cache->table_cache.list, cache.list) {
@@ -2539,8 +2296,12 @@ static int do_list_set(struct netlink_ctx *ctx, struct cmd *cmd,
 static int do_list_hooks(struct netlink_ctx *ctx, struct cmd *cmd)
 {
 	const char *devname = cmd->handle.obj.name;
+	int hooknum = -1;
 
-	return mnl_nft_dump_nf_hooks(ctx, cmd->handle.family, devname);
+	if (cmd->handle.chain.name)
+		hooknum = cmd->handle.chain_id;
+
+	return mnl_nft_dump_nf_hooks(ctx, cmd->handle.family, hooknum, devname);
 }
 
 static int do_command_list(struct netlink_ctx *ctx, struct cmd *cmd)
@@ -2550,16 +2311,10 @@ static int do_command_list(struct netlink_ctx *ctx, struct cmd *cmd)
 	if (nft_output_json(&ctx->nft->output))
 		return do_command_list_json(ctx, cmd);
 
-	if (cmd->handle.table.name != NULL) {
+	if (cmd->handle.table.name != NULL)
 		table = table_cache_find(&ctx->nft->cache.table_cache,
 					 cmd->handle.table.name,
 					 cmd->handle.family);
-		if (!table) {
-			errno = ENOENT;
-			return -1;
-		}
-	}
-
 	switch (cmd->obj) {
 	case CMD_OBJ_TABLE:
 		if (!cmd->handle.table.name)
@@ -2574,6 +2329,7 @@ static int do_command_list(struct netlink_ctx *ctx, struct cmd *cmd)
 	case CMD_OBJ_SET:
 		return do_list_set(ctx, cmd, table);
 	case CMD_OBJ_RULESET:
+	case CMD_OBJ_RULES:
 	case CMD_OBJ_RULE:
 		return do_list_ruleset(ctx, cmd);
 	case CMD_OBJ_METERS:
@@ -2594,10 +2350,8 @@ static int do_command_list(struct netlink_ctx *ctx, struct cmd *cmd)
 	case CMD_OBJ_CT_HELPERS:
 		return do_list_obj(ctx, cmd, NFT_OBJECT_CT_HELPER);
 	case CMD_OBJ_CT_TIMEOUT:
-	case CMD_OBJ_CT_TIMEOUTS:
 		return do_list_obj(ctx, cmd, NFT_OBJECT_CT_TIMEOUT);
 	case CMD_OBJ_CT_EXPECT:
-	case CMD_OBJ_CT_EXPECTATIONS:
 		return do_list_obj(ctx, cmd, NFT_OBJECT_CT_EXPECT);
 	case CMD_OBJ_LIMIT:
 	case CMD_OBJ_LIMITS:
@@ -2614,18 +2368,10 @@ static int do_command_list(struct netlink_ctx *ctx, struct cmd *cmd)
 		return do_list_flowtables(ctx, cmd);
 	case CMD_OBJ_HOOKS:
 		return do_list_hooks(ctx, cmd);
-	case CMD_OBJ_MONITOR:
-	case CMD_OBJ_MARKUP:
-	case CMD_OBJ_SETELEMS:
-	case CMD_OBJ_EXPR:
-	case CMD_OBJ_ELEMENTS:
-		errno = EOPNOTSUPP;
-		return -1;
-	case CMD_OBJ_INVALID:
-		break;
+	default:
+		BUG("invalid command object type %u\n", cmd->obj);
 	}
 
-	BUG("invalid command object type %u\n", cmd->obj);
 	return 0;
 }
 
@@ -2692,6 +2438,14 @@ static int do_command_reset(struct netlink_ctx *ctx, struct cmd *cmd)
 	case CMD_OBJ_QUOTA:
 		type = NFT_OBJECT_QUOTA;
 		break;
+	case CMD_OBJ_RULES:
+		ret = netlink_reset_rules(ctx, cmd, true);
+		if (ret < 0)
+			return ret;
+
+		return do_command_list(ctx, cmd);
+	case CMD_OBJ_RULE:
+		return netlink_reset_rules(ctx, cmd, false);
 	default:
 		BUG("invalid command object type %u\n", cmd->obj);
 	}
@@ -2809,6 +2563,7 @@ int do_command(struct netlink_ctx *ctx, struct cmd *cmd)
 	case CMD_REPLACE:
 		return do_command_replace(ctx, cmd);
 	case CMD_DELETE:
+	case CMD_DESTROY:
 		return do_command_delete(ctx, cmd);
 	case CMD_GET:
 		return do_command_get(ctx, cmd);
@@ -2955,8 +2710,10 @@ static void stmt_reduce(const struct rule *rule)
 
 		/* Must not merge across other statements */
 		if (stmt->ops->type != STMT_EXPRESSION) {
-			if (idx >= 2)
-				payload_do_merge(sa, idx);
+			if (idx < 2)
+				continue;
+
+			payload_do_merge(sa, idx);
 			idx = 0;
 			continue;
 		}
@@ -2970,6 +2727,7 @@ static void stmt_reduce(const struct rule *rule)
 			switch (stmt->expr->op) {
 			case OP_EQ:
 			case OP_IMPLICIT:
+			case OP_NEQ:
 				break;
 			default:
 				continue;
@@ -2980,7 +2738,8 @@ static void stmt_reduce(const struct rule *rule)
 			switch (stmt->expr->op) {
 			case OP_EQ:
 			case OP_IMPLICIT:
-				if (stmt->expr->left->meta.key == NFT_META_PROTOCOL) {
+				if (stmt->expr->left->meta.key == NFT_META_PROTOCOL &&
+				    !stmt->expr->left->meta.inner_desc) {
 					uint16_t protocol;
 
 					protocol = mpz_get_uint16(stmt->expr->right->value);
